import copy
import os
import uuid

import dolphie.Modules.MetricManager as MetricManager
from dolphie.DataTypes import ConnectionSource, ConnectionStatus, Panels
from dolphie.Dolphie import Dolphie
from dolphie.Modules.ArgumentParser import Config, HostGroupMember
from dolphie.Modules.ManualException import ManualException
from dolphie.Modules.ReplayManager import ReplayManager
from dolphie.Widgets.SpinnerWidget import SpinnerWidget
from dolphie.Widgets.TabSetupModal import TabSetupModal
from dolphie.Widgets.TopBar import TopBar
from rich.text import Text
from textual.app import App
from textual.containers import (
    Center,
    Container,
    Horizontal,
    ScrollableContainer,
    VerticalScroll,
)
from textual.content import Content
from textual.timer import Timer
from textual.widgets import (
    Button,
    DataTable,
    Label,
    LoadingIndicator,
    ProgressBar,
    RadioButton,
    RadioSet,
    Sparkline,
    Static,
    Switch,
)
from textual.widgets import Tab as TabWidget
from textual.widgets import TabbedContent, TabPane, Tabs
from textual.worker import Worker


class Tab:
    def __init__(
        self,
        id: str,
        name: str,
        dolphie: Dolphie = None,
        manual_tab_name: str = None,
        replay_manager: ReplayManager = None,
    ):
        self.id = id
        self.name = name
        self.dolphie = dolphie
        self.manual_tab_name = manual_tab_name
        self.replay_manager = replay_manager

        self.worker: Worker = None
        self.worker_timer: Timer = None
        self.worker_cancel_error: ManualException = None

        self.replay_manual_control: bool = False

        self.replicas_worker: Worker = None
        self.replicas_worker_timer: Timer = None

        # Track mounted grid widgets to avoid DOM queries each refresh cycle
        self.clusterset_widgets: dict[str, Static] = {}
        self.galera_widgets: dict[str, Static] = {}
        self.member_widgets: dict[str, Static] = {}
        self.replica_widgets: dict[str, Static] = {}

    def save_references_to_components(self):
        app = self.dolphie.app

        self.main_container = app.query_one("#main_container", VerticalScroll)
        self.metric_graph_tabs = app.query_one("#metric_graph_tabs", TabbedContent)
        self.loading_indicator = app.query_one("#loading_indicator", LoadingIndicator)
        self.sparkline = app.query_one("#panel_dashboard_queries_qps", Sparkline)
        self.panel_dashboard = app.query_one("#panel_dashboard", Container)
        self.panel_graphs = app.query_one("#panel_graphs", Container)
        self.panel_replication = app.query_one("#panel_replication", Container)
        self.panel_metadata_locks = app.query_one("#panel_metadata_locks", Container)
        self.panel_processlist = app.query_one("#panel_processlist", Container)
        self.panel_ddl = app.query_one("#panel_ddl", Container)
        self.panel_pfs_metrics = app.query_one("#panel_pfs_metrics", Container)
        self.panel_statements_summary = app.query_one("#panel_statements_summary", Container)
        self.panel_proxysql_hostgroup_summary = app.query_one("#panel_proxysql_hostgroup_summary", Container)
        self.panel_proxysql_mysql_query_rules = app.query_one("#panel_proxysql_mysql_query_rules", Container)
        self.panel_proxysql_command_stats = app.query_one("#panel_proxysql_command_stats", Container)

        self.spinner = app.query_one("#spinner", SpinnerWidget)
        self.spinner.hide()

        self.ddl_title = app.query_one("#ddl_title", Label)
        self.ddl_datatable = app.query_one("#ddl_datatable", DataTable)

        self.pfs_metrics_file_io_datatable = app.query_one("#pfs_metrics_file_io_datatable", DataTable)
        self.pfs_metrics_table_io_waits_datatable = app.query_one("#pfs_metrics_table_io_waits_datatable", DataTable)
        self.pfs_metrics_radio_set = app.query_one("#pfs_metrics_radio_set", RadioSet)
        self.pfs_metrics_delta = app.query_one("#pfs_metrics_delta", RadioButton)
        self.pfs_metrics_tabs = app.query_one("#pfs_metrics_tabs", TabbedContent)

        self.processlist_title = app.query_one("#processlist_title", Label)
        self.processlist_datatable = app.query_one("#processlist_data", DataTable)
        self.statements_summary_title = app.query_one("#statements_summary_title", Label)
        self.statements_summary_datatable = app.query_one("#statements_summary_datatable", DataTable)
        self.statements_summary_radio_set = app.query_one("#statements_summary_radio_set", RadioSet)
        self.metadata_locks_title = app.query_one("#metadata_locks_title", Label)
        self.metadata_locks_datatable = app.query_one("#metadata_locks_datatable", DataTable)
        self.proxysql_hostgroup_summary_title = app.query_one("#proxysql_hostgroup_summary_title", Static)
        self.proxysql_hostgroup_summary_datatable = app.query_one("#proxysql_hostgroup_summary_datatable", DataTable)
        self.proxysql_mysql_query_rules_title = app.query_one("#proxysql_mysql_query_rules_title", Static)
        self.proxysql_mysql_query_rules_datatable = app.query_one("#proxysql_mysql_query_rules_datatable", DataTable)
        self.proxysql_command_stats_title = app.query_one("#proxysql_command_stats_title", Static)
        self.proxysql_command_stats_datatable = app.query_one("#proxysql_command_stats_datatable", DataTable)

        self.dashboard_replay_container = app.query_one("#dashboard_replay_container", Container)
        self.dashboard_replay_progressbar = app.query_one("#dashboard_replay_progressbar", ProgressBar)
        self.dashboard_replay_start_end = app.query_one("#dashboard_replay_start_end", Static)
        self.dashboard_replay = app.query_one("#dashboard_replay", Static)
        self.dashboard_section_1 = app.query_one("#dashboard_section_1", Static)
        self.dashboard_section_2 = app.query_one("#dashboard_section_2", Static)
        self.dashboard_section_3 = app.query_one("#dashboard_section_3", Static)
        self.dashboard_section_4 = app.query_one("#dashboard_section_4", Static)
        self.dashboard_section_5 = app.query_one("#dashboard_section_5", Static)
        self.dashboard_section_6 = app.query_one("#dashboard_section_6", Static)

        self.clusterset_container = app.query_one("#clusterset_container", Container)
        self.clusterset_title = app.query_one("#clusterset_title", Label)
        self.clusterset_grid = app.query_one("#clusterset_grid", Container)

        self.galera_container = app.query_one("#galera_container", Container)
        self.galera_title = app.query_one("#galera_title", Label)
        self.galera_data = app.query_one("#galera_data", Static)
        self.galera_grid = app.query_one("#galera_grid", Container)

        self.group_replication_container = app.query_one("#group_replication_container", Container)
        self.group_replication_grid = app.query_one("#group_replication_grid", Container)
        self.group_replication_data = app.query_one("#group_replication_data", Static)
        self.group_replication_title = app.query_one("#group_replication_title", Label)

        self.replicas_grid = app.query_one("#replicas_grid", Container)
        self.replicas_container = app.query_one("#replicas_container", Container)
        self.replicas_title = app.query_one("#replicas_title", Label)
        self.replicas_loading_indicator = app.query_one("#replicas_loading_indicator", LoadingIndicator)

        self.replication_container = app.query_one("#replication_container", Container)
        self.replication_variables = app.query_one("#replication_variables", Label)
        self.replication_status = app.query_one("#replication_status", Static)
        self.replication_thread_applier_container = app.query_one(
            "#replication_thread_applier_container", ScrollableContainer
        )
        self.replication_thread_applier = app.query_one("#replication_thread_applier", Static)

    def get_panel_widget(self, panel_name: str) -> Container:
        return getattr(self, f"panel_{panel_name}")

    def refresh_replay_dashboard_section(self):
        if not self.dolphie.replay_file:
            return

        min_timestamp = self.replay_manager.min_replay_timestamp
        max_timestamp = self.replay_manager.max_replay_timestamp
        current_timestamp = self.replay_manager.current_replay_timestamp

        # Highlight if the max timestamp matches the current timestamp
        max_timestamp = (
            f"[b][$green]{max_timestamp}[/b][$green]" if max_timestamp == current_timestamp else max_timestamp
        )

        # Update the dashboard title with the timestamp of the replay event
        self.dashboard_replay.update(
            f"[b]Replay[/b] ([$dark_gray]{os.path.basename(self.dolphie.replay_file)}[/$dark_gray])"
        )
        self.dashboard_replay_start_end.update(
            f"{min_timestamp} [$b_highlight]<-[/$b_highlight] "
            f"[$b_light_blue]{current_timestamp}[/$b_light_blue] [$b_highlight]->[/$b_highlight] "
            f"{max_timestamp}"
        )

        # Update the progress bar with the current replay progress
        if self.replay_manager.current_replay_id == self.replay_manager.min_replay_id:
            current_position = 0
        else:
            current_position = self.replay_manager.current_replay_id - self.replay_manager.min_replay_id + 1

        self.dashboard_replay_progressbar.update(progress=current_position, total=self.replay_manager.total_replay_rows)

    def toggle_entities_displays(self):
        def toggle_tab(tab_name, visible):
            if visible:
                self.metric_graph_tabs.show_tab(tab_name)
            else:
                self.metric_graph_tabs.hide_tab(tab_name)

        self.dashboard_section_6.display = bool(self.dolphie.system_utilization)
        toggle_tab("graph_tab_system", self.dolphie.system_utilization)

        if self.dolphie.connection_source == ConnectionSource.mysql:
            self.dashboard_section_5.display = bool(
                self.dolphie.replication_status and not self.dolphie.panels.replication.visible
            )

            toggle_tab("graph_tab_replication_lag", self.dolphie.replication_status)
            toggle_tab(
                "graph_tab_adaptive_hash_index",
                self.dolphie.global_variables.get("innodb_adaptive_hash_index") != "OFF",
            )
            toggle_tab(
                "graph_tab_locks",
                (self.dolphie.metadata_locks_enabled and self.dolphie.panels.metadata_locks.visible)
                or self.dolphie.replay_file,
            )

        elif self.dolphie.connection_source == ConnectionSource.proxysql:
            self.dashboard_section_5.display = False

    def toggle_metric_graph_tabs_display(self):
        self.main_container.display = True

        # Hide/show the tabs that are available for the current connection source
        for metric_instance in self.dolphie.metric_manager.metrics.__dict__.values():
            if self.dolphie.connection_source in metric_instance.connection_source:
                self.metric_graph_tabs.show_tab(f"graph_tab_{metric_instance.tab_name}")
            else:
                self.metric_graph_tabs.hide_tab(f"graph_tab_{metric_instance.tab_name}")

        # Only show the replay section if we're in replay mode
        self.dashboard_replay_container.display = bool(self.dolphie.replay_file)

        # Update the graph switch values based on the tab's metric data so each tab can have
        # its own set of visible metrics
        metric_switches = self.dolphie.app.tab_manager.metric_switches
        for metric_instance_name, metric_instance in self.dolphie.metric_manager.metrics.__dict__.items():
            for metric, metric_data in metric_instance.__dict__.items():
                if (
                    isinstance(metric_data, MetricManager.MetricData)
                    and metric_data.graphable
                    and metric_data.create_switch
                ):
                    switch = metric_switches.get(f"{metric_instance_name}-{metric}")
                    if switch:
                        switch.value = metric_data.visible

        # Layout redo log graphs based on whether active count is available
        show_active_count = bool(
            self.dolphie.global_status.get("Active_redo_log_count") and not self.dolphie.replay_file
        )
        self.graph_redo_log_data_written.styles.width = "55%" if show_active_count else "88%"
        self.graph_redo_log_active_count.display = show_active_count
        if show_active_count:
            self.dolphie.metric_manager.metrics.redo_log_active_count.Active_redo_log_count.visible = True

    def toggle_replication_panel_components(self):
        def toggle_container_display(container: Container, items, tracked: dict):
            container.display = bool(items)
            for widget in tracked.values():
                widget.parent.display = True

        toggle_container_display(
            self.galera_container, self.dolphie.galera_cluster_members, self.galera_widgets
        )
        toggle_container_display(
            self.replicas_container, self.dolphie.replica_manager.available_replicas, self.replica_widgets
        )
        toggle_container_display(
            self.group_replication_container, self.dolphie.group_replication_members, self.member_widgets
        )
        toggle_container_display(
            self.clusterset_container, self.dolphie.innodb_cluster_clustersets, self.clusterset_widgets
        )

    def remove_replication_panel_components(self):
        for tracked in (self.replica_widgets, self.member_widgets, self.galera_widgets, self.clusterset_widgets):
            for widget in tracked.values():
                widget.parent.remove()
            tracked.clear()



class TabManager:
    def __init__(self, app: App, config: Config):
        self.app = app
        self.config = config

        self.active_tab: Tab = None
        self.tabs: dict[str, Tab] = {}

        self.host_tabs = self.app.query_one("#host_tabs", Tabs)

        self.loading_hostgroups: bool = False
        self.last_replay_time: int = 0

        self.topbar = self.app.query_one(TopBar)

    def update_connection_status(self, tab: Tab, connection_status: ConnectionStatus):
        previous_status = tab.dolphie.connection_status
        tab.dolphie.connection_status = connection_status
        self.update_topbar(tab=tab)

        # Only rename when the host info may have changed (not for read_write/read_only toggles)
        if previous_status not in (ConnectionStatus.read_write, ConnectionStatus.read_only):
            self.rename_tab(tab)

    def update_topbar(self, tab: Tab):
        dolphie = tab.dolphie

        # If we're in daemon mode, don't waste time on this
        if dolphie.daemon_mode:
            return

        # Only update the topbar if we're on the active tab
        if self.active_tab is tab:
            if dolphie.connection_status:
                self.topbar.connection_status = dolphie.connection_status
                self.topbar.host = dolphie.host_with_port

                if (
                    dolphie.record_for_replay
                    and tab.replay_manager
                    and dolphie.connection_status != ConnectionStatus.disconnected
                ):
                    self.topbar.replay_file_size = tab.replay_manager.replay_file_size
                else:
                    self.topbar.replay_file_size = None
            else:
                self.topbar.replay_file_size = None
                self.topbar.connection_status = None
                self.topbar.host = ""

    async def create_ui_widgets(self):
        if self.config.daemon_mode:
            return

        await self.app.mount(
            LoadingIndicator(id="loading_indicator"),
            VerticalScroll(
                SpinnerWidget(id="spinner", text="Processing command"),
                Center(
                    Container(
                        Static(id="dashboard_replay", classes="dashboard_replay"),
                        Static(id="dashboard_replay_start_end", classes="dashboard_replay"),
                        Horizontal(
                            Button("⏪ Back", id="back_button", classes="replay_button"),
                            Button("⏸️  Pause", id="pause_button", classes="replay_button"),
                            Button("⏩ Forward", id="forward_button", classes="replay_button"),
                            Button("🔍 Seek", id="seek_button", classes="replay_button"),
                            classes="replay_buttons",
                        ),
                        ProgressBar(
                            id="dashboard_replay_progressbar", total=100, show_percentage=False, show_eta=False
                        ),
                        id="dashboard_replay_container",
                        classes="dashboard_replay",
                    )
                ),
                Container(
                    Center(
                        Static(id="dashboard_section_1", classes="panel_container"),
                        Static(id="dashboard_section_6", classes="panel_container"),
                        Static(id="dashboard_section_2", classes="panel_container"),
                        Static(id="dashboard_section_3", classes="panel_container"),
                        Static(id="dashboard_section_5", classes="panel_container"),
                        Static(id="dashboard_section_4", classes="panel_container"),
                    ),
                    Sparkline([], id="panel_dashboard_queries_qps"),
                    id="panel_dashboard",
                    classes="dashboard",
                ),
                Container(Label(id="metric_graphs_title", classes="panel_title"), TabbedContent(id="metric_graph_tabs"), id="panel_graphs"),
                Container(
                    Container(
                        Label(id="replication_title", classes="panel_title"),
                        Label(id="replication_variables"),
                        Center(
                            ScrollableContainer(Static(id="replication_status"), classes="replication_status"),
                            ScrollableContainer(
                                Static(id="replication_thread_applier"),
                                id="replication_thread_applier_container",
                                classes="replication_thread_applier",
                            ),
                        ),
                        id="replication_container",
                        classes="replication",
                    ),
                    Container(
                        Label(id="clusterset_title", classes="panel_title"),
                        Container(id="clusterset_grid"),
                        id="clusterset_container",
                        classes="group_replication",
                    ),
                    Container(
                        Label(id="galera_title", classes="panel_title"),
                        Static(id="galera_data"),
                        Container(id="galera_grid"),
                        id="galera_container",
                        classes="group_replication",
                    ),
                    Container(
                        Label(id="group_replication_title", classes="panel_title"),
                        Label(id="group_replication_data"),
                        Container(id="group_replication_grid"),
                        id="group_replication_container",
                        classes="group_replication",
                    ),
                    Container(
                        Label(id="replicas_title", classes="panel_title"),
                        LoadingIndicator(id="replicas_loading_indicator"),
                        Container(id="replicas_grid"),
                        id="replicas_container",
                        classes="replicas",
                    ),
                    id="panel_replication",
                    classes="replication_panel",
                ),
                Container(
                    Label(id="metadata_locks_title"),
                    DataTable(id="metadata_locks_datatable", show_cursor=False, zebra_stripes=True),
                    id="panel_metadata_locks",
                    classes="panel_container",
                ),
                Container(
                    Label(id="ddl_title"),
                    DataTable(id="ddl_datatable", show_cursor=False),
                    id="panel_ddl",
                    classes="panel_container",
                ),
                Container(
                    Label(id="pfs_metrics_title"),
                    RadioSet(
                        *(
                            [
                                RadioButton("Delta since last reset", id="pfs_metrics_delta", value=True),
                                RadioButton("Total since MySQL restart", id="pfs_metrics_total"),
                            ]
                        ),
                        id="pfs_metrics_radio_set",
                    ),
                    TabbedContent(id="pfs_metrics_tabs"),
                    id="panel_pfs_metrics",
                    classes="panel_container",
                ),
                Container(
                    Label(id="proxysql_hostgroup_summary_title"),
                    DataTable(id="proxysql_hostgroup_summary_datatable", show_cursor=False),
                    id="panel_proxysql_hostgroup_summary",
                    classes="panel_container",
                ),
                Container(
                    Label(id="proxysql_mysql_query_rules_title"),
                    DataTable(
                        id="proxysql_mysql_query_rules_datatable",
                        classes="proxysql_mysql_query_rules_datatable",
                        show_cursor=False,
                    ),
                    id="panel_proxysql_mysql_query_rules",
                    classes="panel_container",
                ),
                Container(
                    Label(id="proxysql_command_stats_title"),
                    DataTable(
                        id="proxysql_command_stats_datatable",
                        classes="proxysql_command_stats_datatable",
                        show_cursor=False,
                    ),
                    id="panel_proxysql_command_stats",
                    classes="panel_container",
                ),
                Container(
                    Label(id="processlist_title"),
                    DataTable(id="processlist_data", show_cursor=False),
                    id="panel_processlist",
                    classes="panel_container",
                ),
                Container(
                    Label(id="statements_summary_title"),
                    Label(
                        Text.from_markup(":bulb: [label]Prepared statements are not included in this panel"),
                        id="statements_summary_info",
                    ),
                    RadioSet(
                        *(
                            [
                                RadioButton("Delta since panel opened", id="statements_summarys_delta", value=True),
                                RadioButton("Delta since last sample", id="statements_summary_delta_last_sample"),
                                RadioButton("Total since MySQL restart", id="statements_summary_total"),
                            ]
                        ),
                        id="statements_summary_radio_set",
                    ),
                    DataTable(id="statements_summary_datatable", show_cursor=False),
                    id="panel_statements_summary",
                    classes="panel_container",
                ),
                classes="tab",
                id="main_container",
            ),
        )

        self.app.query_one("#main_container").display = False
        self.app.query_one("#loading_indicator").display = False

        panels = Panels()
        self.app.query_one("#metric_graphs_title", Label).update(panels.graphs.title)
        self.app.query_one("#replication_title", Label).update(panels.replication.title)
        self.app.query_one("#pfs_metrics_title", Label).update(panels.pfs_metrics.title)
        self.app.query_one("#statements_summary_title", Label).update(panels.statements_summary.title)

        # Loop the metric instances and create the graph tabs
        metric_manager = MetricManager.MetricManager(None)
        metric_graph_tabs = self.app.query_one("#metric_graph_tabs", TabbedContent)
        for metric_instance_name, metric_instance in metric_manager.metrics.__dict__.items():
            metric_tab_name = metric_instance.tab_name
            graph_names = metric_instance.graphs
            graph_tab_name = metric_instance.graph_tab_name

            if not self.app.query(f"#graph_tab_{metric_tab_name}"):
                await metric_graph_tabs.add_pane(
                    TabPane(
                        graph_tab_name,
                        Label(id=f"metric_graph_stats_{metric_tab_name}", classes="metric_graph_stats"),
                        Horizontal(id=f"metric_graph_container_{metric_tab_name}", classes="metric_graph_container"),
                        Horizontal(
                            id=f"switch_container_{metric_tab_name}",
                            classes="switch_container switch_container",
                        ),
                        id=f"graph_tab_{metric_tab_name}",
                        name=metric_tab_name,
                    )
                )

            tab_pane = self.app.query_one(f"#graph_tab_{metric_tab_name}", TabPane)
            graph_containers = {}
            for graph_name in graph_names:
                graph_container = (
                    "metric_graph_container2"
                    if graph_name in ["graph_system_network", "graph_system_disk_io"]
                    else "metric_graph_container"
                )
                container_id = f"{graph_container}_{metric_tab_name}"

                # Add graph_container2 only if it's needed
                if container_id not in graph_containers and not self.app.query(f"#{container_id}"):
                    if graph_container == "metric_graph_container2":
                        horizontal = Horizontal(id=container_id, classes="metric_graph_container2")
                        await tab_pane.mount(horizontal, after=1)
                        graph_containers[container_id] = horizontal

                if container_id not in graph_containers:
                    graph_containers[container_id] = self.app.query_one(f"#{container_id}", Horizontal)

                await graph_containers[container_id].mount(
                    MetricManager.Graph(id=f"{graph_name}", classes="panel_data")
                )

            switch_container = self.app.query_one(f"#switch_container_{metric_tab_name}", Horizontal)
            for metric, metric_data in metric_instance.__dict__.items():
                if (
                    isinstance(metric_data, MetricManager.MetricData)
                    and metric_data.graphable
                    and metric_data.create_switch
                    and not self.app.query(f"#switch_container_{metric_tab_name} #{metric_instance_name}-{metric}")
                ):
                    switch = Switch(animate=False, id=f"{metric_instance_name}-{metric}", name=metric_tab_name)
                    await switch_container.mount(Label(metric_data.label), switch)

                    # Toggle the switch if the metric is visible (means to enable it by default)
                    if metric_data.visible:
                        switch.toggle()

        # Add the PFS metrics tabs
        pfs_metrics_tabs = self.app.query_one("#pfs_metrics_tabs", TabbedContent)
        await pfs_metrics_tabs.add_pane(
            TabPane(
                "File I/O",
                DataTable(id="pfs_metrics_file_io_datatable", show_cursor=False),
                id="pfs_metrics_file_io_tab",
            )
        )
        await pfs_metrics_tabs.add_pane(
            TabPane(
                "Table I/O Waits Summary",
                Label(
                    Text.from_markup(":bulb: [label]Format for each metric: Wait time (Operations count)"),
                    id="pfs_metrics_format",
                ),
                DataTable(id="pfs_metrics_table_io_waits_datatable", show_cursor=False),
                id="pfs_metrics_table_io_waits_tab",
            ),
        )

        # Set what marker we use for graphs
        for graph in self.app.query(MetricManager.Graph):
            graph.marker = self.config.graph_marker

        # Cache switch references for fast lookup during tab switches
        self.metric_switches: dict[str, Switch] = {}
        for switch in self.app.query(Switch):
            self.metric_switches[switch.id] = switch

    async def create_tab(
        self, tab_name: str = None, hostgroup_member: HostGroupMember = None, switch_tab: bool = True
    ) -> Tab:
        if len(self.app.screen_stack) > 1:
            return

        tab_id = f"t{uuid.uuid4().hex}"

        # Create a new tab instance
        tab = Tab(id=tab_id, name=tab_name)

        # If we're using hostgroups
        config = copy.deepcopy(self.config)
        if hostgroup_member and self.config.hostgroup_hosts:
            config.replay_file = None
            config.host = hostgroup_member.host
            config.port = hostgroup_member.port
            tab.manual_tab_name = hostgroup_member.tab_title

            # If the hostgroup member has a credential profile, update config with its credentials
            credential_profile_data = self.config.credential_profiles.get(hostgroup_member.credential_profile)
            if credential_profile_data:
                config.credential_profile = hostgroup_member.credential_profile

                if credential_profile_data.user:
                    config.user = credential_profile_data.user
                if credential_profile_data.password:
                    config.password = credential_profile_data.password
                if credential_profile_data.socket:
                    config.socket = credential_profile_data.socket
                if credential_profile_data.ssl:
                    config.ssl = credential_profile_data.ssl

        # Create a new Dolphie instance
        dolphie = Dolphie(config=config, app=self.app)
        dolphie.tab_id = tab_id

        # Set the tab's Dolphie instance
        tab.dolphie = dolphie

        # If we're in daemon mode, stop here since we don't need to
        # do anything else with the UI
        if dolphie.daemon_mode:
            self.active_tab = tab
            self.tabs[tab_id] = tab

            for panel in dolphie.daemon_mode_panels:
                getattr(dolphie.panels, panel).visible = True

            return tab

        tab.save_references_to_components()

        # Create the tab in the UI
        initial_tab_name = "" if hostgroup_member else tab_name
        self.host_tabs.add_tab(TabWidget(initial_tab_name, id=tab_id))

        # Loop the metric instances and save references to the graphs and its labels
        for metric_instance in dolphie.metric_manager.metrics.__dict__.values():
            metric_tab_name = metric_instance.tab_name
            graph_names = metric_instance.graphs

            # Save references graph's labels
            setattr(tab, metric_tab_name, self.app.query_one(f"#metric_graph_stats_{metric_tab_name}"))

            # Save references to the graphs
            for graph_name in graph_names:
                setattr(tab, graph_name, self.app.query_one(f"#{graph_name}"))

        if tab.manual_tab_name:
            self.rename_tab(tab, tab.manual_tab_name)

        tab.replication_container.display = False
        tab.replicas_container.display = False
        tab.galera_container.display = False
        tab.group_replication_container.display = False

        # By default, hide all the panels
        for panel in tab.dolphie.panels.all():
            tab.get_panel_widget(panel).display = False

        # Set panels to be visible for the ones the user specifies
        for panel in dolphie.startup_panels:
            tab.get_panel_widget(panel).display = True
            getattr(dolphie.panels, panel).visible = True

        # Set static graph widths (only needs to happen once per tab)
        tab.graph_redo_log_bar.styles.width = "12%"
        tab.graph_redo_log_active_count.styles.width = "33%"
        tab.graph_adaptive_hash_index.styles.width = "50%"
        tab.graph_adaptive_hash_index_hit_ratio.styles.width = "50%"
        tab.graph_system_cpu.styles.width = "50%"
        tab.graph_system_network.styles.width = "50%"
        tab.graph_system_memory.styles.width = "50%"
        tab.graph_system_disk_io.styles.width = "50%"

        # Set the sparkline data to 0
        tab.sparkline.data = [0]

        self.tabs[tab_id] = tab

        if switch_tab:
            self.switch_tab(tab_id)

        return tab

    async def remove_tab(self, tab: Tab):
        self.host_tabs.remove_tab(tab.id)

    def rename_tab(self, tab: Tab, manual_name: str = None):
        if tab.dolphie.daemon_mode:
            return

        new_name = None
        if not manual_name and not tab.manual_tab_name:
            # host_with_port is the full host:port string, we want to split & truncate it to 24 characters
            host = tab.dolphie.host_with_port.split(":")[0][:24]
            if not host:
                return

            # If the last character isn't a letter or number, remove it
            if not host[-1].isalnum():
                host = host[:-1]

            new_name = f"{host}:[dark_gray]{tab.dolphie.port}"
        elif manual_name:
            new_name = manual_name
        elif tab.manual_tab_name:
            new_name = tab.manual_tab_name

        if new_name:
            tab.name = new_name

            if tab.dolphie.replay_file:
                new_name = f"[b recording][Replay][/b recording] {new_name}"

            self.host_tabs.get_tab(tab.id).label = Content.from_rich_text(new_name, console=self.app.console)

    def switch_tab(self, tab_id: str, set_active: bool = True):
        tab = self.get_tab(tab_id)
        if not tab:
            return

        # Update the active/current tab
        self.active_tab = tab

        # Prevent recursive calls
        if set_active:
            self.host_tabs.active = tab_id

        # Update the topbar
        self.update_topbar(tab=tab)

        tab.main_container.display = bool(tab.dolphie.main_db_connection.is_connected())

    def get_tab(self, id: str) -> Tab:
        return self.tabs.get(id)

    async def disconnect_tab(self, tab: Tab, update_topbar: bool = True):
        for worker in (tab.worker, tab.replicas_worker):
            if worker:
                worker.cancel()
        for timer in (tab.worker_timer, tab.replicas_worker_timer):
            if timer:
                timer.stop()

        tab.worker = tab.worker_timer = None
        tab.replicas_worker = tab.replicas_worker_timer = None

        tab.dolphie.main_db_connection.close()
        tab.dolphie.secondary_db_connection.close()

        tab.dolphie.replica_manager.remove_all_replicas()

        if self.active_tab is tab:
            tab.main_container.display = False
            tab.loading_indicator.display = False

        tab.sparkline.data = [0]
        tab.remove_replication_panel_components()

        if update_topbar:
            self.update_connection_status(tab=tab, connection_status=ConnectionStatus.disconnected)

    def setup_host_tab(self, tab: Tab):
        dolphie = tab.dolphie

        async def command_get_input(data):
            # Set tab_setup to False since it's only used when Dolphie first loads
            if self.config.tab_setup:
                self.config.tab_setup = False

            hostgroup = data.get("hostgroup")
            if hostgroup:
                self.config.record_for_replay = data.get("record_for_replay")
                dolphie.app.connect_as_hostgroup(hostgroup)
            else:
                # Disconnect the existing tab (cancel workers, close connections, cleanup UI)
                await self.disconnect_tab(tab, update_topbar=False)

                # Update connection details on the existing Dolphie instance
                host_port = data["host"].split(":")
                dolphie.host = host_port[0]
                dolphie.port = int(host_port[1]) if len(host_port) > 1 else 3306
                dolphie.credential_profile = data.get("credential_profile")
                dolphie.user = data.get("username")
                dolphie.password = data.get("password")
                dolphie.socket = data.get("socket_file")
                dolphie.ssl = data.get("ssl")
                dolphie.record_for_replay = data.get("record_for_replay")
                dolphie.replay_file = data.get("replay_file")

                # Reset all runtime state with the new connection details
                dolphie.reset_runtime_variables()
                tab.worker_cancel_error = None
                tab.replay_manager = None

                tab.loading_indicator.display = True
                tab.dashboard_replay_container.display = False

                if dolphie.replay_file:
                    tab.replay_manager = ReplayManager(dolphie)
                    if not tab.replay_manager.verify_replay_file():
                        tab.loading_indicator.display = False
                        self.setup_host_tab(tab)
                        return

                    self.update_connection_status(tab=tab, connection_status=ConnectionStatus.connected)
                    dolphie.app.run_worker_replay(tab.id)
                else:
                    dolphie.app.run_worker_main(tab.id)
                    dolphie.app.run_worker_replicas(tab.id)

        # If we're here because of a worker cancel error or manually disconnected,
        # we want to pre-populate the host/port
        if (
            tab.worker_cancel_error
            or dolphie.connection_status == ConnectionStatus.disconnected
            or self.config.tab_setup
        ):
            host = dolphie.host
            port = dolphie.port
        else:
            host = ""
            port = ""

        dolphie.app.push_screen(
            TabSetupModal(
                credential_profile=dolphie.credential_profile,
                credential_profiles=dolphie.config.credential_profiles,
                host=host,
                port=port,
                username=dolphie.user,
                password=dolphie.password,
                ssl=dolphie.ssl,
                record_for_replay=dolphie.record_for_replay,
                socket_file=dolphie.socket,
                hostgroups=dolphie.hostgroup_hosts.keys(),
                available_hosts=dolphie.tab_setup_available_hosts,
                replay_directory=dolphie.config.replay_dir,
                replay_files=dolphie.get_replay_files(),
                error_message=tab.worker_cancel_error,
            ),
            command_get_input,
        )



# Use __future__ to allow 'deque[int]' type hint in MetricData
from __future__ import annotations

import contextlib
import dataclasses
from collections import defaultdict, deque
from collections.abc import Callable
from dataclasses import dataclass, field
from datetime import datetime, timedelta
from enum import Enum
from typing import Union

import plotext as plt
from rich.text import Text
from textual.widgets import Static

from dolphie.DataTypes import ConnectionSource
from dolphie.Modules.Functions import format_bytes, format_number, format_time


class MetricSource(Enum):
    """Enumeration of sources for metric data."""

    SYSTEM_UTILIZATION = "system_utilization"
    GLOBAL_STATUS = "global_status"
    INNODB_METRICS = "innodb_metrics"
    DISK_IO_METRICS = "disk_io_metrics"
    PROXYSQL_SELECT_COMMAND_STATS = "proxysql_select_command_stats"
    PROXYSQL_TOTAL_COMMAND_STATS = "proxysql_total_command_stats"
    NONE = "none"


@dataclass
class MetricColor:
    """Namespace for standard metric graph colors."""

    gray: tuple = (172, 207, 231)
    blue: tuple = (68, 180, 255)
    green: tuple = (84, 239, 174)
    red: tuple = (255, 73, 112)
    yellow: tuple = (252, 213, 121)
    purple: tuple = (191, 121, 252)
    orange: tuple = (252, 121, 121)


class Graph(Static):
    """A Textual widget for rendering time-series graphs using plotext."""

    def __init__(self, *args, **kwargs) -> None:
        """Initialize the Graph widget."""
        super().__init__(*args, **kwargs)
        self.marker: str | None = None
        self.metric_instance: MetricInstance | None = None
        # This will be a deque, but we treat it as an iterable
        self.datetimes: deque[str] | None = None

    def on_resize(self) -> None:
        """Re-render the graph when the widget is resized."""
        self.render_graph(self.metric_instance, self.datetimes)

    def _setup_plot(self) -> None:
        """Clears and configures the plotext canvas."""
        plt.clf()
        plt.date_form("d/m/y H:M:S")
        plt.canvas_color((10, 14, 27))
        plt.axes_color((10, 14, 27))
        plt.ticks_color((133, 159, 213))
        plt.plotsize(self.size.width, self.size.height)

    def _finalize_plot(self, max_y_value: float) -> None:
        """Calculates Y-ticks, formats labels, and updates the widget."""
        max_y_ticks = 5
        y_tick_interval = (max_y_value / max_y_ticks) if max_y_ticks > 0 else 0

        if y_tick_interval >= 1:
            y_ticks = [i * y_tick_interval for i in range(int(max_y_ticks) + 1)]
        else:
            y_ticks = [float(i) for i in range(int(max_y_value) + 2)]

        format_function = get_number_format_function(self.metric_instance)
        y_labels = [format_function(val) for val in y_ticks]

        plt.yticks(y_ticks, y_labels)

        with contextlib.suppress(OSError):
            self.update(Text.from_ansi(plt.build()))

    def _render_checkpoint_metrics(self, x: list[str], y: list[float]) -> float:
        """Renders the graph for CheckpointMetrics."""
        plt.hline(0, (10, 14, 27))
        plt.hline(self.metric_instance.checkpoint_age_sync_flush, (241, 251, 130))
        plt.hline(self.metric_instance.checkpoint_age_max, (252, 121, 121))

        max_x = max(x)
        plt.text(
            "Critical",
            y=self.metric_instance.checkpoint_age_max,
            x=max_x,
            alignment="right",
            color=(233, 233, 233),
            style="bold",
        )
        plt.text(
            "Warning",
            y=self.metric_instance.checkpoint_age_sync_flush,
            x=max_x,
            alignment="right",
            color=(233, 233, 233),
            style="bold",
        )

        metric = self.metric_instance.Innodb_checkpoint_age
        plt.plot(x, y, marker=self.marker, label=metric.label, color=metric.color)
        return self.metric_instance.checkpoint_age_max

    def _render_redo_log_bar_metrics(self, y_values: list[int]) -> float:
        """Renders the bar graph for RedoLogMetrics."""
        x = [0]
        # Calculate y from the snapshot of values
        y = [round(sum(y_values) * (3600 / len(y_values)))]

        log_size = self.metric_instance.redo_log_size
        plt.hline(log_size, (252, 121, 121))
        plt.text(
            "Log Size",
            y=log_size,
            x=0,
            alignment="center",
            color=(233, 233, 233),
            style="bold",
        )

        bar_color = (46, 124, 175)
        if y[0] >= log_size:
            bar_color = (252, 121, 121)

        plt.text(
            format_bytes(y[0], color=False) + "/hr",
            y=y[0],
            x=0,
            alignment="center",
            color=(233, 233, 233),
            style="bold",
            background=bar_color,
        )
        plt.bar(x, y, marker="hd", color=bar_color)
        return max(log_size, max(y))

    def _render_redo_log_line_metrics(self, x: list[str], y: list[float]) -> float:
        """Renders the line graph for RedoLogMetrics."""
        metric = self.metric_instance.Innodb_lsn_current
        plt.plot(x, y, marker=self.marker, label=metric.label, color=metric.color)
        return max(y) if y else 0

    def _render_active_redo_log_metrics(self, x: list[str], y: list[float]) -> float:
        """Renders the graph for RedoLogActiveCountMetrics."""
        plt.hline(1, (10, 14, 27))
        plt.hline(34, (252, 121, 121))
        plt.text(
            "Max Count",
            y=34,
            x=max(x),
            alignment="right",
            color=(233, 233, 233),
            style="bold",
        )

        metric = self.metric_instance.Active_redo_log_count
        plt.plot(x, y, marker=self.marker, label=metric.label, color=metric.color)
        return 34.0  # Fixed max Y for this graph

    def _render_system_memory_metrics(self, x: list[str], y: list[float]) -> float:
        """Renders the graph for SystemMemoryMetrics."""
        total_mem = self.metric_instance.Memory_Total.last_value or 0
        plt.hline(0, (10, 14, 27))
        plt.hline(total_mem, (252, 121, 121))
        plt.text(
            "Total",
            y=total_mem,
            x=max(x),
            alignment="right",
            color=(233, 233, 233),
            style="bold",
        )

        metric = self.metric_instance.Memory_Used
        plt.plot(x, y, marker=self.marker, label=metric.label, color=metric.color)
        return total_mem

    def _render_default_metrics(self, x: list[str]) -> float:
        """Renders a graph for any standard metric instance."""
        max_y = 0.0
        for metric_data in self.metric_instance.__dict__.values():
            if isinstance(metric_data, MetricData) and metric_data.visible:
                # **THREAD-SAFETY**: Snapshot deque to list
                y = list(metric_data.values)
                if y and x:
                    plt.plot(
                        x,
                        y,
                        marker=self.marker,
                        label=metric_data.label,
                        color=metric_data.color,
                    )
                    try:
                        max_y = max(max_y, max(y))
                    except ValueError:
                        pass  # Handle empty list
        return max_y

    def render_graph(self, metric_instance: MetricInstance | None, datetimes: deque[str] | None) -> None:
        """Renders a graph for the given metric instance and datetimes.

        Args:
            metric_instance: The metric dataclass instance to plot.
            datetimes: A deque of datetime strings for the X-axis.
        """
        self.metric_instance = metric_instance
        self.datetimes = datetimes

        if self.metric_instance is None or self.datetimes is None:
            self.update("")  # Clear the graph if no data
            return

        self._setup_plot()

        max_y_value = 0.0

        # Create a snapshot of the datetimes and all
        # relevant metric values at the beginning of the render for thread-safety
        try:
            x = list(self.datetimes)
            if not x:  # Check if list is empty after copy
                self.update("")
                return
        except RuntimeError:  # deque changed size during iteration
            self.update("")
            return

        try:
            if isinstance(self.metric_instance, CheckpointMetrics):
                y = list(self.metric_instance.Innodb_checkpoint_age.values)
                if x and y:
                    max_y_value = self._render_checkpoint_metrics(x, y)

            elif isinstance(self.metric_instance, RedoLogMetrics):
                if "graph_redo_log_bar" in self.id:
                    # Snapshot the values needed for the bar chart
                    innodb_lsn_values = list(self.metric_instance.Innodb_lsn_current.values)
                    if innodb_lsn_values:
                        max_y_value = self._render_redo_log_bar_metrics(innodb_lsn_values)
                else:
                    y = list(self.metric_instance.Innodb_lsn_current.values)
                    if x and y:
                        max_y_value = max(max_y_value, self._render_redo_log_line_metrics(x, y))

            elif isinstance(self.metric_instance, RedoLogActiveCountMetrics):
                y = list(self.metric_instance.Active_redo_log_count.values)
                if x and y:
                    max_y_value = self._render_active_redo_log_metrics(x, y)

            elif isinstance(self.metric_instance, SystemMemoryMetrics):
                y = list(self.metric_instance.Memory_Used.values)
                if x and y:
                    max_y_value = self._render_system_memory_metrics(x, y)

            else:
                # Default renderer snapshots its own 'y' values inside
                max_y_value = self._render_default_metrics(x)

        except (ValueError, TypeError, IndexError):
            pass  # Catch errors during plotting

        self._finalize_plot(max_y_value)


def get_number_format_function(data: MetricInstance, color: bool = False) -> Callable[[int | float], str]:
    """Returns the correct formatting function based on the metric type."""
    data_formatters: dict[type, Callable[[int | float], str]] = {
        ReplicationLagMetrics: lambda val: format_time(val),
        CheckpointMetrics: lambda val: format_bytes(val, color=color),
        RedoLogMetrics: lambda val: format_bytes(val, color=color),
        AdaptiveHashIndexHitRatio: lambda val: f"{round(val)}%",
        ProxySQLMultiplexEfficiency: lambda val: f"{round(val)}%",
        DiskIOMetrics: lambda val: format_bytes(val, color=color),
        ProxySQLQueriesDataNetwork: lambda val: format_bytes(val, color=color),
        SystemMemoryMetrics: lambda val: format_bytes(val, color=color),
        SystemNetworkMetrics: lambda val: format_bytes(val, color=color),
    }
    return data_formatters.get(type(data), lambda val: format_number(val, color=color))


@dataclass
class MetricData:
    label: str
    color: tuple
    visible: bool = True
    save_history: bool = True
    per_second_calculation: bool = True
    last_value: int | None = None
    graphable: bool = True
    create_switch: bool = True
    # Use a deque for O(1) appends and pops
    values: deque[int] = field(default_factory=deque)


@dataclass
class SystemCPUMetrics:
    CPU_Percent: MetricData
    graphs: list[str]
    tab_name: str = "system"
    graph_tab_name = "System"
    metric_source: MetricSource = MetricSource.SYSTEM_UTILIZATION
    connection_source: list[ConnectionSource] = field(
        default_factory=lambda: [ConnectionSource.mysql, ConnectionSource.proxysql]
    )
    use_with_replay: bool = True


@dataclass
class SystemMemoryMetrics:
    Memory_Total: MetricData
    Memory_Used: MetricData
    graphs: list[str]
    tab_name: str = "system"
    graph_tab_name = "System"
    metric_source: MetricSource = MetricSource.SYSTEM_UTILIZATION
    connection_source: list[ConnectionSource] = field(
        default_factory=lambda: [ConnectionSource.mysql, ConnectionSource.proxysql]
    )
    use_with_replay: bool = True


@dataclass
class SystemNetworkMetrics:
    Network_Down: MetricData
    Network_Up: MetricData
    graphs: list[str]
    tab_name: str = "system"
    graph_tab_name = "System"
    metric_source: MetricSource = MetricSource.SYSTEM_UTILIZATION
    connection_source: list[ConnectionSource] = field(
        default_factory=lambda: [ConnectionSource.mysql, ConnectionSource.proxysql]
    )
    use_with_replay: bool = True


@dataclass
class SystemDiskIOMetrics:
    Disk_Read: MetricData
    Disk_Write: MetricData
    graphs: list[str]
    tab_name: str = "system"
    graph_tab_name = "System"
    metric_source: MetricSource = MetricSource.SYSTEM_UTILIZATION
    connection_source: list[ConnectionSource] = field(
        default_factory=lambda: [ConnectionSource.mysql, ConnectionSource.proxysql]
    )
    use_with_replay: bool = True


@dataclass
class DMLMetrics:
    Queries: MetricData
    Com_select: MetricData
    Com_insert: MetricData
    Com_update: MetricData
    Com_delete: MetricData
    Com_replace: MetricData
    Com_commit: MetricData
    Com_rollback: MetricData
    graphs: list[str]
    tab_name: str = "dml"
    graph_tab_name = "DML"
    metric_source: MetricSource = MetricSource.GLOBAL_STATUS
    connection_source: list[ConnectionSource] = field(
        default_factory=lambda: [ConnectionSource.mysql, ConnectionSource.proxysql]
    )
    use_with_replay: bool = True


@dataclass
class ReplicationLagMetrics:
    lag: MetricData
    graphs: list[str]
    tab_name: str = "replication_lag"
    graph_tab_name = "Replication"
    metric_source: MetricSource = MetricSource.NONE
    connection_source: list[ConnectionSource] = field(default_factory=lambda: [ConnectionSource.mysql])
    use_with_replay: bool = True


@dataclass
class CheckpointMetrics:
    Innodb_checkpoint_age: MetricData
    graphs: list[str]
    tab_name: str = "checkpoint"
    graph_tab_name = "Checkpoint"
    metric_source: MetricSource = MetricSource.GLOBAL_STATUS
    checkpoint_age_max: int = 0
    checkpoint_age_sync_flush: int = 0
    connection_source: list[ConnectionSource] = field(default_factory=lambda: [ConnectionSource.mysql])
    use_with_replay: bool = True


@dataclass
class BufferPoolRequestsMetrics:
    Innodb_buffer_pool_read_requests: MetricData
    Innodb_buffer_pool_write_requests: MetricData
    Innodb_buffer_pool_reads: MetricData
    graphs: list[str]
    tab_name: str = "buffer_pool_requests"
    graph_tab_name = "BP Requests"
    metric_source: MetricSource = MetricSource.GLOBAL_STATUS
    connection_source: list[ConnectionSource] = field(default_factory=lambda: [ConnectionSource.mysql])
    use_with_replay: bool = True


@dataclass
class AdaptiveHashIndexMetrics:
    adaptive_hash_searches: MetricData
    adaptive_hash_searches_btree: MetricData
    graphs: list[str]
    tab_name: str = "adaptive_hash_index"
    graph_tab_name = "AHI"
    metric_source: MetricSource = MetricSource.INNODB_METRICS
    connection_source: list[ConnectionSource] = field(default_factory=lambda: [ConnectionSource.mysql])
    use_with_replay: bool = True


@dataclass
class AdaptiveHashIndexHitRatio:
    hit_ratio: MetricData
    graphs: list[str]
    smoothed_hit_ratio: float | None = None
    tab_name: str = "adaptive_hash_index"
    graph_tab_name = "AHI"
    metric_source: MetricSource = MetricSource.NONE
    connection_source: list[ConnectionSource] = field(default_factory=lambda: [ConnectionSource.mysql])
    use_with_replay: bool = True


@dataclass
class RedoLogMetrics:
    Innodb_lsn_current: MetricData
    graphs: list[str]
    tab_name: str = "redo_log"
    graph_tab_name = "Redo Log"
    redo_log_size: int = 0
    metric_source: MetricSource = MetricSource.GLOBAL_STATUS
    connection_source: list[ConnectionSource] = field(default_factory=lambda: [ConnectionSource.mysql])
    use_with_replay: bool = True


@dataclass
class RedoLogActiveCountMetrics:
    Active_redo_log_count: MetricData
    graphs: list[str]
    tab_name: str = "redo_log"
    graph_tab_name = "Redo Log"
    metric_source: MetricSource = MetricSource.GLOBAL_STATUS
    connection_source: list[ConnectionSource] = field(default_factory=lambda: [ConnectionSource.mysql])
    use_with_replay: bool = True


@dataclass
class TableCacheMetrics:
    Table_open_cache_hits: MetricData
    Table_open_cache_misses: MetricData
    Table_open_cache_overflows: MetricData
    graphs: list[str]
    tab_name: str = "table_cache"
    graph_tab_name = "Table Cache"
    metric_source: MetricSource = MetricSource.GLOBAL_STATUS
    connection_source: list[ConnectionSource] = field(default_factory=lambda: [ConnectionSource.mysql])
    use_with_replay: bool = True


@dataclass
class ThreadMetrics:
    Threads_connected: MetricData
    Threads_running: MetricData
    graphs: list[str]
    tab_name: str = "threads"
    graph_tab_name = "Threads"
    metric_source: MetricSource = MetricSource.GLOBAL_STATUS
    connection_source: list[ConnectionSource] = field(default_factory=lambda: [ConnectionSource.mysql])
    use_with_replay: bool = True


@dataclass
class TemporaryObjectMetrics:
    Created_tmp_tables: MetricData
    Created_tmp_disk_tables: MetricData
    Created_tmp_files: MetricData
    graphs: list[str]
    tab_name: str = "temporary_objects"
    graph_tab_name = "Temp Objects"
    metric_source: MetricSource = MetricSource.GLOBAL_STATUS
    connection_source: list[ConnectionSource] = field(default_factory=lambda: [ConnectionSource.mysql])
    use_with_replay: bool = True


@dataclass
class AbortedConnectionsMetrics:
    Aborted_clients: MetricData
    Aborted_connects: MetricData
    graphs: list[str]
    tab_name: str = "aborted_connections"
    graph_tab_name = "Aborted Connections"
    metric_source: MetricSource = MetricSource.GLOBAL_STATUS
    connection_source: list[ConnectionSource] = field(default_factory=lambda: [ConnectionSource.mysql])
    use_with_replay: bool = True


@dataclass
class DiskIOMetrics:
    io_read: MetricData
    io_write: MetricData
    graphs: list[str]
    tab_name: str = "disk_io"
    graph_tab_name = "Disk I/O"
    metric_source: MetricSource = MetricSource.DISK_IO_METRICS
    connection_source: list[ConnectionSource] = field(default_factory=lambda: [ConnectionSource.mysql])
    use_with_replay: bool = True


@dataclass
class LocksMetrics:
    metadata_lock_count: MetricData
    graphs: list[str]
    tab_name: str = "locks"
    graph_tab_name = "Locks"
    metric_source: MetricSource = MetricSource.NONE
    connection_source: list[ConnectionSource] = field(default_factory=lambda: [ConnectionSource.mysql])
    use_with_replay: bool = True


@dataclass
class HistoryListLength:
    trx_rseg_history_len: MetricData
    graphs: list[str]
    tab_name: str = "history_list_length"
    graph_tab_name = "History List"
    metric_source: MetricSource = MetricSource.INNODB_METRICS
    connection_source: list[ConnectionSource] = field(default_factory=lambda: [ConnectionSource.mysql])
    use_with_replay: bool = True


@dataclass
class ProxySQLConnectionsMetrics:
    Client_Connections_non_idle: MetricData
    Client_Connections_aborted: MetricData
    Client_Connections_connected: MetricData
    Client_Connections_created: MetricData
    Server_Connections_aborted: MetricData
    Server_Connections_connected: MetricData
    Server_Connections_created: MetricData
    Access_Denied_Wrong_Password: MetricData
    graphs: list[str]
    tab_name: str = "proxysql_connections"
    graph_tab_name = "Connections"
    metric_source: MetricSource = MetricSource.GLOBAL_STATUS
    connection_source: list[ConnectionSource] = field(default_factory=lambda: [ConnectionSource.proxysql])
    use_with_replay: bool = True


@dataclass
class ProxySQLQueriesDataNetwork:
    Queries_backends_bytes_recv: MetricData
    Queries_backends_bytes_sent: MetricData
    Queries_frontends_bytes_recv: MetricData
    Queries_frontends_bytes_sent: MetricData
    graphs: list[str]
    tab_name: str = "proxysql_queries_data_network"
    graph_tab_name = "Query Data Rates"
    metric_source: MetricSource = MetricSource.GLOBAL_STATUS
    connection_source: list[ConnectionSource] = field(default_factory=lambda: [ConnectionSource.proxysql])
    use_with_replay: bool = True


@dataclass
class ProxySQLActiveTRX:
    Active_Transactions: MetricData
    graphs: list[str]
    tab_name: str = "proxysql_active_trx"
    graph_tab_name = "Active TRX"
    metric_source: MetricSource = MetricSource.GLOBAL_STATUS
    connection_source: list[ConnectionSource] = field(default_factory=lambda: [ConnectionSource.proxysql])
    use_with_replay: bool = True


@dataclass
class ProxySQLMultiplexEfficiency:
    proxysql_multiplex_efficiency_ratio: MetricData
    graphs: list[str]
    tab_name: str = "proxysql_multiplex_efficiency"
    graph_tab_name = "Multiplex Efficiency"
    metric_source: MetricSource = MetricSource.GLOBAL_STATUS
    connection_source: list[ConnectionSource] = field(default_factory=lambda: [ConnectionSource.proxysql])
    use_with_replay: bool = True


@dataclass
class ProxySQLSELECTCommandStats:
    cnt_100us: MetricData
    cnt_500us: MetricData
    cnt_1ms: MetricData
    cnt_5ms: MetricData
    cnt_10ms: MetricData
    cnt_50ms: MetricData
    cnt_100ms: MetricData
    cnt_500ms: MetricData
    cnt_1s: MetricData
    cnt_5s: MetricData
    cnt_10s: MetricData
    cnt_INFs: MetricData
    graphs: list[str]
    tab_name: str = "proxysql_select_command_stats"
    graph_tab_name = "SELECT Command Stats"
    metric_source: MetricSource = MetricSource.PROXYSQL_SELECT_COMMAND_STATS
    connection_source: list[ConnectionSource] = field(default_factory=lambda: [ConnectionSource.proxysql])
    use_with_replay: bool = True


@dataclass
class ProxySQLTotalCommandStats:
    cnt_100us: MetricData
    cnt_500us: MetricData
    cnt_1ms: MetricData
    cnt_5ms: MetricData
    cnt_10ms: MetricData
    cnt_50ms: MetricData
    cnt_100ms: MetricData
    cnt_500ms: MetricData
    cnt_1s: MetricData
    cnt_5s: MetricData
    cnt_10s: MetricData
    cnt_INFs: MetricData
    graphs: list[str]
    tab_name: str = "proxysql_total_command_stats"
    graph_tab_name = "Total Command Stats"
    metric_source: MetricSource = MetricSource.PROXYSQL_TOTAL_COMMAND_STATS
    connection_source: list[ConnectionSource] = field(default_factory=lambda: [ConnectionSource.proxysql])
    use_with_replay: bool = True


# Type alias for all metric types
MetricInstance = Union[
    SystemCPUMetrics,
    SystemMemoryMetrics,
    SystemNetworkMetrics,
    SystemDiskIOMetrics,
    DMLMetrics,
    ReplicationLagMetrics,
    CheckpointMetrics,
    BufferPoolRequestsMetrics,
    AdaptiveHashIndexMetrics,
    AdaptiveHashIndexHitRatio,
    RedoLogMetrics,
    RedoLogActiveCountMetrics,
    TableCacheMetrics,
    ThreadMetrics,
    TemporaryObjectMetrics,
    AbortedConnectionsMetrics,
    DiskIOMetrics,
    LocksMetrics,
    HistoryListLength,
    ProxySQLConnectionsMetrics,
    ProxySQLQueriesDataNetwork,
    ProxySQLActiveTRX,
    ProxySQLMultiplexEfficiency,
    ProxySQLSELECTCommandStats,
    ProxySQLTotalCommandStats,
]


@dataclass
class MetricInstances:
    """Container for all specific metric instances."""

    system_cpu: SystemCPUMetrics
    system_memory: SystemMemoryMetrics
    system_disk_io: SystemDiskIOMetrics
    system_network: SystemNetworkMetrics
    dml: DMLMetrics
    buffer_pool_requests: BufferPoolRequestsMetrics
    history_list_length: HistoryListLength
    adaptive_hash_index: AdaptiveHashIndexMetrics
    adaptive_hash_index_hit_ratio: AdaptiveHashIndexHitRatio
    checkpoint: CheckpointMetrics
    redo_log_active_count: RedoLogActiveCountMetrics
    redo_log: RedoLogMetrics
    table_cache: TableCacheMetrics
    threads: ThreadMetrics
    temporary_objects: TemporaryObjectMetrics
    aborted_connections: AbortedConnectionsMetrics
    disk_io: DiskIOMetrics
    locks: LocksMetrics
    replication_lag: ReplicationLagMetrics
    proxysql_active_trx: ProxySQLActiveTRX
    proxysql_multiplex_efficiency: ProxySQLMultiplexEfficiency
    proxysql_connections: ProxySQLConnectionsMetrics
    proxysql_queries_data_network: ProxySQLQueriesDataNetwork
    proxysql_select_command_stats: ProxySQLSELECTCommandStats
    proxysql_total_command_stats: ProxySQLTotalCommandStats


class MetricManager:
    """Manages the state, collection, and processing of all metrics."""

    DATETIME_FORMAT = "%d/%m/%y %H:%M:%S"
    ROLLING_WINDOW_MINUTES = 10

    def __init__(self, replay_file: str, daemon_mode: bool = False):
        """Initialize the MetricManager.

        Args:
            replay_file: Path to a replay file, if one is being used.
            daemon_mode: True if running in daemon mode (trims old data).
        """
        self.connection_source = ConnectionSource.mysql
        self.replay_file = replay_file
        self.daemon_mode = daemon_mode

        # Attributes populated by refresh_data
        self.worker_start_time: datetime | None = None
        self.system_utilization: dict[str, int] = {}
        self.innodb_metrics: dict[str, int] = {}
        self.disk_io_metrics: dict[str, int] = {}
        self.metadata_lock_metrics: dict[str, int] = {}
        self.replication_status: dict[str, int | str] = {}
        self.proxysql_total_command_stats: dict[str, int] = {}
        self.proxysql_select_command_stats: dict[str, int] = {}

        # State attributes
        self.initialized: bool = False
        self.polling_latency: float = 0
        self.global_variables: dict[str, int | str] = {}
        self.global_status: dict[str, int] = {}
        self.redo_log_size: int = 0
        # Use a deque for O(1) appends and pops
        self.datetimes: deque[str] = deque()

        # The authoritative structure of all metrics
        self.metrics: MetricInstances = None  # type: ignore

        # Optimized lookup tables for processing
        # For fast, source-based processing in update_..._values
        self._source_to_metrics_processing: dict[MetricSource, list[tuple[str, MetricData, list[ConnectionSource]]]] = (
            defaultdict(list)
        )
        # For fast history cleanup in daemon_cleanup_data
        self._all_metrics_data_history: list[MetricData] = []

        # Setup the dispatch map for metric sources
        self._metric_source_map: dict[MetricSource, dict[str, int] | None] = {
            MetricSource.SYSTEM_UTILIZATION: self.system_utilization,
            MetricSource.GLOBAL_STATUS: self.global_status,
            MetricSource.INNODB_METRICS: self.innodb_metrics,
            MetricSource.DISK_IO_METRICS: self.disk_io_metrics,
            MetricSource.PROXYSQL_SELECT_COMMAND_STATS: self.proxysql_select_command_stats,
            MetricSource.PROXYSQL_TOTAL_COMMAND_STATS: self.proxysql_total_command_stats,
            MetricSource.NONE: None,
        }

        self.reset()

    def reset(self):
        """Resets all metrics and state to their default values."""
        self.initialized = False
        self.polling_latency = 0
        self.global_variables.clear()
        self.global_status.clear()
        self.redo_log_size = 0
        self.datetimes.clear()

        # Clear raw data stores
        self.system_utilization.clear()
        self.innodb_metrics.clear()
        self.disk_io_metrics.clear()
        self.metadata_lock_metrics.clear()
        self.replication_status.clear()
        self.proxysql_total_command_stats.clear()
        self.proxysql_select_command_stats.clear()

        # Clear performance lookup tables
        self._source_to_metrics_processing.clear()
        self._all_metrics_data_history.clear()

        self.metrics = MetricInstances(
            system_cpu=SystemCPUMetrics(
                graphs=["graph_system_cpu"],
                CPU_Percent=MetricData(
                    label="CPU %",
                    color=MetricColor.blue,
                    per_second_calculation=False,
                    create_switch=False,
                ),
            ),
            system_memory=SystemMemoryMetrics(
                graphs=["graph_system_memory"],
                Memory_Total=MetricData(
                    label="Total",
                    color=MetricColor.blue,
                    per_second_calculation=False,
                    visible=False,
                    save_history=False,
                    create_switch=False,
                ),
                Memory_Used=MetricData(
                    label="Memory Used",
                    color=MetricColor.green,
                    per_second_calculation=False,
                    create_switch=False,
                ),
            ),
            system_disk_io=SystemDiskIOMetrics(
                graphs=["graph_system_disk_io"],
                Disk_Read=MetricData(label="IOPS Read", color=MetricColor.blue),
                Disk_Write=MetricData(label="IOPS Write", color=MetricColor.yellow),
            ),
            system_network=SystemNetworkMetrics(
                graphs=["graph_system_network"],
                Network_Down=MetricData(label="Net Dn", color=MetricColor.blue),
                Network_Up=MetricData(label="Net Up", color=MetricColor.gray),
            ),
            dml=DMLMetrics(
                graphs=["graph_dml"],
                Queries=MetricData(label="Queries", color=MetricColor.gray, visible=False),
                Com_select=MetricData(label="SELECT", color=MetricColor.blue),
                Com_insert=MetricData(label="INSERT", color=MetricColor.green),
                Com_update=MetricData(label="UPDATE", color=MetricColor.yellow),
                Com_delete=MetricData(label="DELETE", color=MetricColor.red),
                Com_replace=MetricData(
                    label="REPLACE",
                    color=MetricColor.red,
                    visible=False,
                    save_history=False,
                    graphable=False,
                ),
                Com_commit=MetricData(
                    label="COMMIT",
                    color=MetricColor.green,
                    visible=False,
                    save_history=True,
                    graphable=False,
                ),
                Com_rollback=MetricData(
                    label="ROLLBACK",
                    color=MetricColor.red,
                    visible=False,
                    save_history=False,
                    graphable=False,
                ),
            ),
            replication_lag=ReplicationLagMetrics(
                graphs=["graph_replication_lag"],
                lag=MetricData(
                    label="Lag",
                    color=MetricColor.blue,
                    per_second_calculation=False,
                    create_switch=False,
                ),
            ),
            checkpoint=CheckpointMetrics(
                graphs=["graph_checkpoint"],
                Innodb_checkpoint_age=MetricData(
                    label="Uncheckpointed",
                    color=MetricColor.green,
                    per_second_calculation=False,
                    create_switch=False,
                ),
            ),
            buffer_pool_requests=BufferPoolRequestsMetrics(
                graphs=["graph_buffer_pool_requests"],
                Innodb_buffer_pool_read_requests=MetricData(label="Read Requests", color=MetricColor.blue),
                Innodb_buffer_pool_write_requests=MetricData(label="Write Requests", color=MetricColor.green),
                Innodb_buffer_pool_reads=MetricData(label="Disk Reads", color=MetricColor.red),
            ),
            history_list_length=HistoryListLength(
                graphs=["graph_history_list_length"],
                trx_rseg_history_len=MetricData(
                    label="HLL",
                    color=MetricColor.blue,
                    per_second_calculation=False,
                    create_switch=False,
                ),
            ),
            adaptive_hash_index=AdaptiveHashIndexMetrics(
                graphs=["graph_adaptive_hash_index"],
                adaptive_hash_searches=MetricData(label="Hit", color=MetricColor.green),
                adaptive_hash_searches_btree=MetricData(label="Miss", color=MetricColor.red),
            ),
            adaptive_hash_index_hit_ratio=AdaptiveHashIndexHitRatio(
                graphs=["graph_adaptive_hash_index_hit_ratio"],
                hit_ratio=MetricData(
                    label="Hit Ratio",
                    color=MetricColor.green,
                    per_second_calculation=False,
                    create_switch=False,
                ),
            ),
            redo_log=RedoLogMetrics(
                graphs=["graph_redo_log_data_written", "graph_redo_log_bar"],
                Innodb_lsn_current=MetricData(label="Data Written", color=MetricColor.blue, create_switch=False),
            ),
            redo_log_active_count=RedoLogActiveCountMetrics(
                graphs=["graph_redo_log_active_count"],
                Active_redo_log_count=MetricData(
                    label="Active Count",
                    color=MetricColor.blue,
                    per_second_calculation=False,
                    visible=False,
                    create_switch=False,
                ),
            ),
            table_cache=TableCacheMetrics(
                graphs=["graph_table_cache"],
                Table_open_cache_hits=MetricData(label="Hit", color=MetricColor.green),
                Table_open_cache_misses=MetricData(label="Miss", color=MetricColor.red),
                Table_open_cache_overflows=MetricData(label="Overflow", color=MetricColor.yellow),
            ),
            threads=ThreadMetrics(
                graphs=["graph_threads"],
                Threads_connected=MetricData(
                    label="Connected",
                    color=MetricColor.green,
                    per_second_calculation=False,
                    visible=False,
                ),
                Threads_running=MetricData(
                    label="Running",
                    color=MetricColor.blue,
                    per_second_calculation=False,
                ),
            ),
            temporary_objects=TemporaryObjectMetrics(
                graphs=["graph_temporary_objects"],
                Created_tmp_tables=MetricData(label="Tables", color=MetricColor.blue),
                Created_tmp_disk_tables=MetricData(label="Disk", color=MetricColor.red),
                Created_tmp_files=MetricData(label="Files", color=MetricColor.yellow),
            ),
            aborted_connections=AbortedConnectionsMetrics(
                graphs=["graph_aborted_connections"],
                Aborted_clients=MetricData(label="Client (timeout)", color=MetricColor.blue),
                Aborted_connects=MetricData(label="Connects (attempt)", color=MetricColor.red),
            ),
            disk_io=DiskIOMetrics(
                graphs=["graph_disk_io"],
                io_read=MetricData(label="Read", color=MetricColor.blue),
                io_write=MetricData(label="Write", color=MetricColor.yellow),
            ),
            locks=LocksMetrics(
                graphs=["graph_locks"],
                metadata_lock_count=MetricData(
                    label="Metadata",
                    color=MetricColor.red,
                    per_second_calculation=False,
                ),
            ),
            proxysql_connections=ProxySQLConnectionsMetrics(
                graphs=["graph_proxysql_connections"],
                Client_Connections_aborted=MetricData(label="FE (aborted)", color=MetricColor.gray),
                Client_Connections_connected=MetricData(
                    label="FE (connected)",
                    color=MetricColor.green,
                    per_second_calculation=False,
                    visible=False,
                ),
                Client_Connections_created=MetricData(label="FE (created)", color=MetricColor.yellow),
                Server_Connections_aborted=MetricData(label="BE (aborted)", color=MetricColor.red),
                Server_Connections_connected=MetricData(
                    label="BE (connected)",
                    color=MetricColor.green,
                    per_second_calculation=False,
                    visible=False,
                ),
                Server_Connections_created=MetricData(label="BE (created)", color=MetricColor.blue),
                Access_Denied_Wrong_Password=MetricData(label="Wrong Password", color=MetricColor.purple),
                Client_Connections_non_idle=MetricData(
                    label="FE (non-idle)",
                    color=MetricColor.green,
                    per_second_calculation=False,
                    visible=True,
                ),
            ),
            proxysql_queries_data_network=ProxySQLQueriesDataNetwork(
                graphs=["graph_proxysql_queries_data_network"],
                Queries_backends_bytes_recv=MetricData(label="BE Recv", color=MetricColor.blue),
                Queries_backends_bytes_sent=MetricData(label="BE Sent", color=MetricColor.green),
                Queries_frontends_bytes_recv=MetricData(label="FE Recv", color=MetricColor.purple),
                Queries_frontends_bytes_sent=MetricData(label="FE Sent", color=MetricColor.yellow),
            ),
            proxysql_active_trx=ProxySQLActiveTRX(
                graphs=["graph_proxysql_active_trx"],
                Active_Transactions=MetricData(
                    label="Active TRX",
                    color=MetricColor.blue,
                    per_second_calculation=False,
                    create_switch=False,
                ),
            ),
            proxysql_multiplex_efficiency=ProxySQLMultiplexEfficiency(
                graphs=["graph_proxysql_multiplex_efficiency"],
                proxysql_multiplex_efficiency_ratio=MetricData(
                    label="Multiplex Efficiency",
                    color=MetricColor.blue,
                    per_second_calculation=False,
                    create_switch=False,
                ),
            ),
            proxysql_select_command_stats=ProxySQLSELECTCommandStats(
                graphs=["graph_proxysql_select_command_stats"],
                cnt_100us=MetricData(label="100us", color=MetricColor.gray, visible=False),
                cnt_500us=MetricData(label="500us", color=MetricColor.blue, visible=False),
                cnt_1ms=MetricData(label="1ms", color=MetricColor.green, visible=False),
                cnt_5ms=MetricData(label="5ms", color=MetricColor.green, visible=False),
                cnt_10ms=MetricData(label="10ms", color=MetricColor.green),
                cnt_50ms=MetricData(label="50ms", color=MetricColor.yellow),
                cnt_100ms=MetricData(label="100ms", color=MetricColor.yellow),
                cnt_500ms=MetricData(label="500ms", color=MetricColor.orange),
                cnt_1s=MetricData(label="1s", color=MetricColor.orange),
                cnt_5s=MetricData(label="5s", color=MetricColor.red),
                cnt_10s=MetricData(label="10s", color=MetricColor.purple),
                cnt_INFs=MetricData(label="10s+", color=MetricColor.purple),
            ),
            proxysql_total_command_stats=ProxySQLTotalCommandStats(
                graphs=["graph_proxysql_total_command_stats"],
                cnt_100us=MetricData(label="100us", color=MetricColor.gray, visible=False),
                cnt_500us=MetricData(label="500us", color=MetricColor.blue, visible=False),
                cnt_1ms=MetricData(label="1ms", color=MetricColor.green, visible=False),
                cnt_5ms=MetricData(label="5ms", color=MetricColor.green, visible=False),
                cnt_10ms=MetricData(label="10ms", color=MetricColor.green),
                cnt_50ms=MetricData(label="50ms", color=MetricColor.yellow),
                cnt_100ms=MetricData(label="100ms", color=MetricColor.yellow),
                cnt_500ms=MetricData(label="500ms", color=MetricColor.orange),
                cnt_1s=MetricData(label="1s", color=MetricColor.orange),
                cnt_5s=MetricData(label="5s", color=MetricColor.red),
                cnt_10s=MetricData(label="10s", color=MetricColor.purple),
                cnt_INFs=MetricData(label="10s+", color=MetricColor.purple),
            ),
        )

        # Build the optimized lookup tables
        for metric_instance in self.metrics.__dict__.values():
            if not dataclasses.is_dataclass(metric_instance):
                continue

            source = getattr(metric_instance, "metric_source", MetricSource.NONE)
            conn_source = getattr(metric_instance, "connection_source", [])

            for attr_name, metric_data in metric_instance.__dict__.items():
                if isinstance(metric_data, MetricData):
                    if metric_data.save_history:
                        self._all_metrics_data_history.append(metric_data)

                    # Add to processing list if it has a valid source
                    if source != MetricSource.NONE:
                        self._source_to_metrics_processing[source].append((attr_name, metric_data, conn_source))

    def refresh_data(
        self,
        worker_start_time: datetime,
        polling_latency: float = 0,
        system_utilization: dict[str, int] = None,
        global_variables: dict[str, int | str] = None,
        global_status: dict[str, int] = None,
        innodb_metrics: dict[str, int] = None,
        proxysql_command_stats: list[dict[str, str]] = None,
        disk_io_metrics: dict[str, int] = None,
        metadata_lock_metrics: dict[str, int] = None,
        replication_status: dict[str, int | str] = None,
    ):
        """Ingests new data from a polling worker and updates all metric values."""
        if replication_status is None:
            replication_status = {}
        if metadata_lock_metrics is None:
            metadata_lock_metrics = {}
        if disk_io_metrics is None:
            disk_io_metrics = {}
        if proxysql_command_stats is None:
            proxysql_command_stats = []
        if innodb_metrics is None:
            innodb_metrics = {}
        if global_status is None:
            global_status = {}
        if global_variables is None:
            global_variables = {}
        if system_utilization is None:
            system_utilization = {}
        self.worker_start_time = worker_start_time
        self.polling_latency = polling_latency
        self.system_utilization.update(system_utilization)
        self.global_variables = global_variables
        self.global_status.update(global_status)
        self.innodb_metrics.update(innodb_metrics)
        self.disk_io_metrics.update(disk_io_metrics)
        self.metadata_lock_metrics = metadata_lock_metrics
        self.replication_status = replication_status

        self.proxysql_total_command_stats.clear()
        self.proxysql_select_command_stats.clear()

        # Calculate redo log size
        innodb_redo_log_capacity = self.global_variables.get("innodb_redo_log_capacity", 0)
        innodb_log_file_size = round(
            self.global_variables.get("innodb_log_file_size", 0)
            * self.global_variables.get("innodb_log_files_in_group", 1)
        )
        self.redo_log_size = max(int(innodb_redo_log_capacity), int(innodb_log_file_size))

        if not self.replay_file:
            self.update_proxysql_command_stats(proxysql_command_stats)
            self.update_metrics_per_second_values()
            self.update_metrics_replication_lag()
            self.update_metrics_adaptive_hash_index_hit_ratio()
            self.update_metrics_locks()
            self.update_metrics_last_value()  # Must be last

        self.update_metrics_checkpoint()
        self.metrics.redo_log.redo_log_size = self.redo_log_size

        self.add_metric_datetime()

        if self.daemon_mode:
            self.trim_datetimes_to_window(worker_start_time)

        self.initialized = True

    def add_metric(self, metric_data: MetricData, value: int):
        """Adds a new data point to a metric's value list."""
        if self.initialized:
            if metric_data.save_history:
                metric_data.values.append(value)
            else:
                # If not saving history, just keep the latest value
                if metric_data.values:
                    metric_data.values[0] = value
                else:
                    metric_data.values.append(value)

    def add_metric_datetime(self):
        """Adds the current worker timestamp to the global datetime list."""
        if self.initialized and not self.replay_file and self.worker_start_time:
            self.datetimes.append(self.worker_start_time.strftime(self.DATETIME_FORMAT))

    def get_metric_source_data(self, metric_source: MetricSource) -> dict[str, int] | None:
        """Retrieves the raw data dictionary for a given MetricSource."""
        return self._metric_source_map.get(metric_source)

    def update_metrics_per_second_values(self):
        """Iterates over all metrics and calculates their new per-second values
        using the optimized lookup table.
        """
        for source, metric_tuples in self._source_to_metrics_processing.items():
            metric_source_data = self.get_metric_source_data(source)
            if metric_source_data is None:
                continue

            for metric_name, metric_data, conn_source in metric_tuples:
                if self.connection_source not in conn_source:
                    continue

                current_metric_source_value = metric_source_data.get(metric_name, 0)

                if metric_data.last_value is None:
                    metric_data.last_value = current_metric_source_value
                    continue

                if metric_data.per_second_calculation:
                    metric_diff = current_metric_source_value - metric_data.last_value
                    metric_status_per_sec = round(metric_diff / self.polling_latency) if self.polling_latency > 0 else 0
                else:
                    metric_status_per_sec = current_metric_source_value

                # Special case for CPU_Percent smoothing
                if metric_name == "CPU_Percent":
                    if len(metric_data.values) == 1 and metric_data.values[0] == 0:
                        metric_data.values[0] = metric_status_per_sec
                    elif (
                        metric_status_per_sec in {0, 100}
                        and abs(metric_status_per_sec - (metric_data.values[-1] if metric_data.values else 0)) > 10
                    ):
                        recent_values = list(metric_data.values)[-3:]
                        if recent_values:
                            metric_status_per_sec = sum(recent_values) / len(recent_values)

                self.add_metric(metric_data, int(metric_status_per_sec))

    def update_metrics_last_value(self):
        """Updates the 'last_value' for all metrics using the optimized lookup table."""
        for source, metric_tuples in self._source_to_metrics_processing.items():
            metric_source_data = self.get_metric_source_data(source)
            if metric_source_data is None:
                continue

            for metric_name, metric_data, _ in metric_tuples:
                metric_data.last_value = metric_source_data.get(metric_name, 0)

    def update_proxysql_command_stats(self, proxysql_command_stats: list[dict[str, str]]):
        """Parses and aggregates ProxySQL command stats."""
        if self.connection_source != ConnectionSource.proxysql:
            return

        for row in proxysql_command_stats:
            if row.get("Command") == "SELECT":
                self.proxysql_select_command_stats = {key: int(value) for key, value in row.items() if value.isdigit()}

            for key, value in row.items():
                if key.startswith("cnt_") and value.isdigit():
                    int_value = int(value)
                    self.proxysql_total_command_stats[key] = self.proxysql_total_command_stats.get(key, 0) + int_value

    def update_metrics_replication_lag(self):
        """Updates the replication lag metric."""
        self.add_metric(
            self.metrics.replication_lag.lag,
            int(self.replication_status.get("Seconds_Behind", 0)),
        )

    def update_metrics_adaptive_hash_index_hit_ratio(self):
        """Updates the AHI hit ratio metric from its calculated value."""
        hit_ratio = self.calculate_ahi_ratio()
        if hit_ratio is not None:
            self.add_metric(self.metrics.adaptive_hash_index_hit_ratio.hit_ratio, int(hit_ratio))

    def update_metrics_checkpoint(self):
        """Updates the checkpoint metric instance with max/sync flush values."""
        (max_age, sync_flush, _) = self.calculate_checkpoint_age_data()
        self.metrics.checkpoint.checkpoint_age_max = max_age
        self.metrics.checkpoint.checkpoint_age_sync_flush = sync_flush

    def update_metrics_locks(self):
        """Updates the metadata lock count metric."""
        self.add_metric(self.metrics.locks.metadata_lock_count, len(self.metadata_lock_metrics))

    def calculate_checkpoint_age_data(self) -> tuple[int, int, int]:
        """Calculates raw checkpoint age data."""
        current_age = round(self.global_status.get("Innodb_checkpoint_age", 0))
        max_age = self.redo_log_size

        if max_age == 0:
            return 0, 0, 0

        sync_flush_age = round(max_age * 0.825)
        return max_age, sync_flush_age, current_age

    def get_formatted_checkpoint_age(self) -> str:
        """Gets a color-formatted string for the checkpoint age percentage."""
        (max_age, _, current_age) = self.calculate_checkpoint_age_data()

        if current_age == 0 or max_age == 0:
            return "N/A"

        checkpoint_age_ratio = round(current_age / max_age * 100, 2)
        color_code = "red" if checkpoint_age_ratio >= 80 else "yellow" if checkpoint_age_ratio >= 60 else "green"
        return f"[{color_code}]{checkpoint_age_ratio}%"

    def calculate_ahi_ratio(self) -> float | None:
        """Calculates the smoothed Adaptive Hash Index hit ratio."""
        if self.global_variables.get("innodb_adaptive_hash_index") == "OFF":
            return None

        current_hits = self.innodb_metrics.get("adaptive_hash_searches", 0)
        current_misses = self.innodb_metrics.get("adaptive_hash_searches_btree", 0)

        last_hits = self.metrics.adaptive_hash_index.adaptive_hash_searches.last_value
        last_misses = self.metrics.adaptive_hash_index.adaptive_hash_searches_btree.last_value

        if last_hits is None or last_misses is None:
            return None

        hits = current_hits - last_hits
        misses = current_misses - last_misses
        total_hits_misses = hits + misses

        if total_hits_misses <= 0:
            return 0.0

        hit_ratio = (hits / total_hits_misses) * 100
        smoothing_factor = 0.5
        smoothed_hit_ratio = self.metrics.adaptive_hash_index_hit_ratio.smoothed_hit_ratio

        if smoothed_hit_ratio is None:
            smoothed_hit_ratio = hit_ratio
        else:
            smoothed_hit_ratio = (1 - smoothing_factor) * smoothed_hit_ratio + smoothing_factor * hit_ratio

        self.metrics.adaptive_hash_index_hit_ratio.smoothed_hit_ratio = smoothed_hit_ratio
        return smoothed_hit_ratio

    def get_formatted_ahi_status(self) -> str:
        """Gets a color-formatted string for the AHI status."""
        if self.global_variables.get("innodb_adaptive_hash_index") == "OFF":
            return "OFF"

        smoothed_hit_ratio: float | None = None
        if self.replay_file:
            if self.metrics.adaptive_hash_index_hit_ratio.hit_ratio.values:
                smoothed_hit_ratio = self.metrics.adaptive_hash_index_hit_ratio.hit_ratio.values[-1]
        else:
            smoothed_hit_ratio = self.metrics.adaptive_hash_index_hit_ratio.smoothed_hit_ratio

        if smoothed_hit_ratio is None:
            return "N/A"
        if smoothed_hit_ratio <= 0.01:
            return "Inactive"

        color_code = "green" if smoothed_hit_ratio > 70 else "yellow" if smoothed_hit_ratio > 50 else "red"
        return f"[{color_code}]{smoothed_hit_ratio:.2f}%[/{color_code}]"

    def trim_datetimes_to_window(self, reference_time: datetime) -> bool:
        """Trims datetimes and metric values to the rolling window ending at reference_time.

        Args:
            reference_time: The reference time to calculate the window from.

        Returns:
            bool: True if any entries were trimmed, False otherwise.
        """
        if not self.datetimes:
            return False

        threshold = reference_time.replace(tzinfo=None) - timedelta(minutes=self.ROLLING_WINDOW_MINUTES)
        trimmed = False

        while self.datetimes:
            try:
                first_dt = datetime.strptime(self.datetimes[0], self.DATETIME_FORMAT)
                if first_dt < threshold:
                    self.datetimes.popleft()
                    for metric_data in self._all_metrics_data_history:
                        if metric_data.values:
                            metric_data.values.popleft()
                    trimmed = True
                else:
                    break
            except (ValueError, IndexError):
                try:
                    self.datetimes.popleft()
                    for metric_data in self._all_metrics_data_history:
                        if metric_data.values:
                            metric_data.values.popleft()
                except IndexError:
                    break

        return trimmed




import textwrap

from textual import events, on, work
from textual.app import ComposeResult
from textual.binding import Binding
from textual.containers import Container, Horizontal
from textual.screen import Screen
from textual.widgets import DataTable, Input, Label, Switch

from dolphie.Modules.MySQL import Database
from dolphie.Modules.Queries import MySQLQueries
from dolphie.Widgets.SpinnerWidget import SpinnerWidget
from dolphie.Widgets.TopBar import TopBar


class EventLog(Screen):
    CSS = """
        EventLog {
            & Horizontal {
                height: auto;
                align: center top;
                background: #0a0e1b;
                width: 100%;

                & > Label {
                    color: #bbc8e8;
                    text-style: bold;
                    margin-right: -1;
                }
            }
            & DataTable {
                background: #0a0e1b;
                border: none;
                overflow-x: auto;
                max-height: 100%;

                &:focus {
                    background-tint: #0a0e1b;
                }
            }
            & SpinnerWidget {
                margin-top: 1;
            }
            & .input_container {
                align: left top;
                padding-left: 1;

                & > Input {
                    border: none;
                    background: #0a0e1b;
                    margin: 0;
                    height: 1;
                }
            }
            & #days_container {
                & > Input {
                    width: 15;
                }
                & > Label {
                    margin-right: 2;
                }
            }
            & #info {
                padding-top: 1;
                width: 100%;
                text-align: center;
                text-style: bold;
            }
            & #search {
                width: 90%;
                margin-bottom: 1;
            }
            & #help {
                color: #8f9fc1;
                width: 100%;
                content-align: right middle;
            }
        }
    """

    BINDINGS = [
        Binding("q", "app.pop_screen", "", show=False),
    ]

    def __init__(self, connection_status, app_version, host, db_connection: Database):
        super().__init__()

        self.connection_status = connection_status
        self.app_version = app_version
        self.host = host
        self.db_connection = db_connection

        self.levels = {
            "system": {"active": True, "sql": "prio = 'System'"},
            "warning": {"active": True, "sql": "prio = 'Warning'"},
            "error": {"active": True, "sql": "prio = 'Error'"},
            "note": {"active": True, "sql": "prio = 'Note'"},
        }

    def on_mount(self):
        self.datatable = self.query_one(DataTable)
        self.datatable.focus()

        self.spinner = self.query_one(SpinnerWidget)
        self.info = self.query_one("#info", Label)
        self.search_text = self.query_one("#search", Input)
        self.days_to_display = self.query_one("#days", Input)

        self.info.display = False
        self.datatable.display = False

        self.update_datatable()

    @on(events.Key)
    def on_keypress(self, event: events.Key):
        if event.key == "1":
            self.datatable.move_cursor(row=0)
        elif event.key == "2":
            self.datatable.move_cursor(row=self.datatable.row_count - 1)
        elif event.key == "r":
            self.update_datatable()

    def compose(self) -> ComposeResult:
        yield TopBar(connection_status=self.connection_status, app_version=self.app_version, host=self.host)
        yield Label(
            "[b white]r[/b white] = refresh/[b white]1[/b white] = top of events/"
            "[b white]2[/b white] = bottom of events",
            id="help",
        )
        with Horizontal():
            switch_options = [("System", "system"), ("Warning", "warning"), ("Error", "error"), ("Note", "note")]
            for label, switch_id in switch_options:
                yield Label(label)
                yield Switch(animate=False, id=switch_id, value=True)
        with Horizontal(id="days_container", classes="input_container"):
            yield Label("Days to display")
            yield Input(id="days", value="30")
        with Horizontal(id="search_container", classes="input_container"):
            yield Label("Search event text")
            yield Input(id="search", placeholder="Specify event text to display")
        yield SpinnerWidget(id="spinner", text="Loading events")
        yield Label("", id="info")
        with Container():
            yield DataTable(show_cursor=False)

    @on(Input.Submitted, "Input")
    def event_search(self):
        self.update_datatable()

    @work(thread=True)
    def update_datatable(self):
        for switch in self.query(Switch):
            self.levels[switch.id]["active"] = switch.value

        # Verify days is a number
        try:
            int(self.days_to_display.value)
        except ValueError:
            self.datatable.display = False
            self.info.display = True
            self.info.update("[red]Days to display must be a number[/red]")
            return

        self.spinner.show()

        self.info.display = False
        self.datatable.display = False

        active_sql_list = [data["sql"] for data in self.levels.values() if data["active"]]
        where_clause = " OR ".join(active_sql_list)

        if self.search_text.value:
            where_clause = f"({where_clause}) AND (data LIKE '%{self.search_text.value}%')"

        self.datatable.clear(columns=True)
        self.datatable.add_column("Date/Time")
        self.datatable.add_column("Subsystem")
        self.datatable.add_column("Level")
        self.datatable.add_column("Code")

        if where_clause:
            query = MySQLQueries.error_log.replace("$1", f"AND ({where_clause})")
            query = query.replace("$2", f"AND logged > NOW() - INTERVAL {self.days_to_display.value} DAY")
            event_count = self.db_connection.execute(query)
            data = self.db_connection.fetchall()

            if data:
                self.datatable.add_column(f"Event ({event_count})")

                for row in data:
                    level_color = ""
                    if row["level"] == "Error":
                        level_color = "red"
                    elif row["level"] == "Warning":
                        level_color = "yellow"
                    elif row["level"] == "Note":
                        level_color = "dark_gray"

                    level = row["level"]
                    if level_color:
                        level = f"[{level_color}]{row['level']}[/{level_color}]"

                    timestamp = f"[#858A97]{row['timestamp'].strftime('%Y-%m-%d %H:%M:%S')}[/#858A97]"
                    error_code = f"[label]{row['error_code']}[/label]"
                    subsystem = row["subsystem"]

                    # Wrap the message to 78% of console width so hopefully we don't get a scrollbar
                    wrapped_message = textwrap.wrap(row["message"], width=round(self.app.console.width * 0.75))
                    wrapped_message = "\n".join(wrapped_message)

                    line_counts = [cell.count("\n") + 1 for cell in wrapped_message]
                    height = max(line_counts)

                    self.datatable.add_row(timestamp, subsystem, level, error_code, wrapped_message, height=height)

                self.datatable.display = True
                self.datatable.focus()
            else:
                self.datatable.display = False
                self.info.display = True
                self.info.update("No events found")
        else:
            self.datatable.display = False
            self.info.display = True
            self.info.update("No switches selected. Toggle the switches above to filter what events you'd like to see")

        self.spinner.hide()



from dolphie.DataTypes import ConnectionSource
from dolphie.Modules.Functions import format_number, format_picoseconds, format_query
from dolphie.Modules.TabManager import Tab
from rich.syntax import Syntax

MAX_ROWS = 100


def create_panel(tab: Tab):
    dolphie = tab.dolphie
    datatable = tab.statements_summary_datatable
    query_length_max = 300

    columns = [
        {"name": "Schema", "field": "schema_name", "width": 14, "format_number": False},
        {"name": "Count", "field": "count_star", "width": 8, "format_number": True},
        {
            "name": "Latency",
            "field": "sum_timer_wait",
            "width": 10,
            "format_number": False,
        },
    ]

    if dolphie.is_mysql_version_at_least("8.0") and dolphie.connection_source_alt != ConnectionSource.mariadb:
        columns.extend(
            [
                {
                    "name": "95th %",
                    "field": "quantile_95",
                    "width": 10,
                    "format_number": False,
                },
                {
                    "name": "99th %",
                    "field": "quantile_99",
                    "width": 10,
                    "format_number": False,
                },
            ]
        )

    columns.extend(
        [
            {
                "name": "Lock time",
                "field": "sum_lock_time",
                "width": 9,
                "format_number": False,
            },
            {
                "name": "Rows examined",
                "field": "sum_rows_examined",
                "width": 13,
                "format_number": True,
            },
            {
                "name": "Rows affected",
                "field": "sum_rows_affected",
                "width": 13,
                "format_number": True,
            },
            {
                "name": "Rows sent",
                "field": "sum_rows_sent",
                "width": 9,
                "format_number": True,
            },
            {
                "name": "Errors",
                "field": "sum_errors",
                "width": 6,
                "format_number": True,
            },
            {
                "name": "Warnings",
                "field": "sum_warnings",
                "width": 8,
                "format_number": True,
            },
            {
                "name": "Bad idx",
                "field": "sum_no_good_index_used",
                "width": 7,
                "format_number": True,
            },
            {
                "name": "No idx",
                "field": "sum_no_index_used",
                "width": 6,
                "format_number": True,
            },
            {"name": "Query", "field": "query", "width": None, "format_number": False},
            {
                "name": "latency_total",
                "field": "sum_timer_wait",
                "width": 0,
                "format_number": False,
            },
        ]
    )

    if not dolphie.statements_summary_data or not dolphie.statements_summary_data.filtered_data:
        datatable.display = False
        tab.statements_summary_title.update(
            f"{dolphie.panels.statements_summary.title} " f"([$highlight]0[/$highlight])"
        )

        return

    datatable.display = True

    if len(datatable.columns) != len(columns):
        datatable.clear(columns=True)

    column_names = []
    column_fields = []
    column_format_numbers = []

    if not datatable.columns:
        for column_data in columns:
            datatable.add_column(column_data["name"], key=column_data["name"], width=column_data["width"])

    for column_data in columns:
        column_names.append(column_data["name"])
        column_fields.append(column_data["field"])
        column_format_numbers.append(column_data["format_number"])

    data = dolphie.statements_summary_data.filtered_data
    display_mode = ""
    if tab.statements_summary_radio_set.pressed_button.id == "statements_summary_total":
        display_mode = "t"
    elif tab.statements_summary_radio_set.pressed_button.id == "statements_summarys_delta":
        display_mode = "d"
    else:
        display_mode = "d_last_sample"

    # We sort by sum_timer_wait and retain only the top MAX_ROWS elements to minimize useless
    # UI processing of many rows - at the expense of the sort & slice operations.
    display_data = dict(
        sorted(
            data.items(),
            key=lambda element: element[1]["sum_timer_wait"][display_mode],
            reverse=True,
        )[: MAX_ROWS + 1]
    )

    changed = False

    with dolphie.app.batch_update():
        # Remove stale rows first
        if display_data:
            current_rows = set(display_data.keys())
            existing_rows = set(datatable.rows.keys())

            rows_to_remove = existing_rows - current_rows
            if rows_to_remove:
                changed = True
                if len(rows_to_remove) > len(display_data):
                    datatable.clear()
                else:
                    for row_id in rows_to_remove:
                        datatable.remove_row(row_id)
        else:
            if datatable.row_count:
                changed = True
                datatable.clear()

        for digest, metrics in display_data.items():
            row_values = []
            for column_name, column_field, column_format_number in zip(
                column_names, column_fields, column_format_numbers
            ):
                column_value = metrics.get(column_field, {})

                if isinstance(column_value, dict):
                    column_value = column_value.get(display_mode, 0)

                if column_name == "Query":
                    if dolphie.show_statements_summary_query_digest_text_sample:
                        column_value = metrics.get("query_sample_text")
                    else:
                        column_value = metrics.get("digest_text")
                    column_value = format_query(column_value)
                    # Truncate query Syntax objects
                    if isinstance(column_value, Syntax):
                        column_value = format_query(column_value.code[:query_length_max])
                elif column_name == "Schema":
                    column_value = column_value or "[dark_gray]N/A"
                elif column_format_number:
                    column_value = format_number(column_value)
                elif column_name in ("Latency", "Lock time", "CPU time", "95th %", "99th %"):
                    column_value = format_picoseconds(column_value)

                if column_name != "latency_total" and (column_value == "0" or column_value == 0):
                    column_value = "[dark_gray]0"

                row_values.append(column_value)

            if digest in datatable.rows:
                datatable_row = datatable.get_row(digest)

                for column_id, column_name in enumerate(column_names):
                    new_val = row_values[column_id]
                    old_val = datatable_row[column_id]

                    # Compare text content for Syntax objects
                    cmp_new = new_val.code if isinstance(new_val, Syntax) else new_val
                    cmp_old = old_val.code if isinstance(old_val, Syntax) else old_val

                    if cmp_new != cmp_old:
                        changed = True
                        datatable.update_cell(
                            digest,
                            column_name,
                            new_val,
                            update_width=(column_name == "Query"),
                        )
            else:
                changed = True
                datatable.add_row(*row_values, key=digest)

        if changed:
            datatable.sort("latency_total", reverse=dolphie.sort_by_time_descending)

    title = (
        f"{dolphie.panels.statements_summary.title} "
        f"([$highlight]{datatable.row_count}[/$highlight])"
    )
    tab.statements_summary_title.update(title)



from dolphie.DataTypes import ConnectionSource


class CommandManager:
    def __init__(self):
        self.command_keys = {
            ConnectionSource.mysql: {
                "Commands": {
                    "1": {"human_key": "1", "description": "Toggle panel Dashboard"},
                    "2": {"human_key": "2", "description": "Toggle panel Processlist"},
                    "3": {
                        "human_key": "3",
                        "description": "Toggle panel Metric Graphs",
                    },
                    "4": {
                        "human_key": "4",
                        "description": "Toggle panel Replication/Replicas",
                    },
                    "5": {
                        "human_key": "5",
                        "description": "Toggle panel Metadata Locks",
                    },
                    "6": {"human_key": "6", "description": "Toggle panel DDLs"},
                    "7": {
                        "human_key": "7",
                        "description": "Toggle panel Performance Schema Metrics",
                    },
                    "8": {
                        "human_key": "8",
                        "description": "Toggle panel Statements Summary Metrics",
                    },
                    "placeholder_1": {"human_key": "", "description": ""},
                    "grave_accent": {"human_key": "`", "description": "Open tab setup"},
                    "plus": {"human_key": "+", "description": "Create a new tab"},
                    "minus": {
                        "human_key": "-",
                        "description": "Remove the current tab",
                    },
                    "equals_sign": {
                        "human_key": "=",
                        "description": "Rename the current tab",
                    },
                    "D": {
                        "human_key": "D",
                        "description": "Disconnect from the tab's host",
                    },
                    "ctrl+a": {
                        "human_key": "ctrl+a",
                        "description": "Switch to the previous tab",
                    },
                    "ctrl+d": {
                        "human_key": "ctrl+d",
                        "description": "Switch to the next tab",
                    },
                    "placeholder_2": {"human_key": "", "description": ""},
                    "a": {
                        "human_key": "a",
                        "description": "Toggle additional processlist columns",
                    },
                    "A": {
                        "human_key": "A",
                        "description": "Toggle query digest text/sample text in Statements Summary panel",
                    },
                    "C": {
                        "human_key": "C",
                        "description": "Toggle display of threads with concurrency tickets in Processlist panel",
                    },
                    "i": {
                        "human_key": "i",
                        "description": "Toggle display of idle threads",
                    },
                    "T": {
                        "human_key": "T",
                        "description": "Toggle display of threads that only have an active transaction",
                    },
                    "p": {
                        "human_key": "p",
                        "description": "Toggle pause for refreshing of panels",
                    },
                    "P": {
                        "human_key": "P",
                        "description": (
                            "Toggle between Information Schema and Performance Schema for the Processlist panel"
                        ),
                    },
                    "s": {
                        "human_key": "s",
                        "description": "Toggle sorting for Age in Processlist panel",
                    },
                    "placeholder_3": {"human_key": "", "description": ""},
                    "l": {
                        "human_key": "l",
                        "description": "Display the most recent deadlock",
                    },
                    "o": {
                        "human_key": "o",
                        "description": "Display output from SHOW ENGINE INNODB STATUS",
                    },
                    "d": {"human_key": "d", "description": "Display all databases"},
                    "e": {
                        "human_key": "e",
                        "description": "Display error log from Performance Schema",
                    },
                    "t": {
                        "human_key": "t",
                        "description": "Display details of a thread along with an EXPLAIN of its query",
                    },
                    "u": {
                        "human_key": "u",
                        "description": "Display active connected users and their statistics",
                    },
                    "v": {
                        "human_key": "v",
                        "description": "Display variables from SHOW GLOBAL VARIABLES",
                    },
                    "z": {
                        "human_key": "z",
                        "description": "Display all entries in the host cache",
                    },
                    "Z": {
                        "human_key": "Z",
                        "description": (
                            "Display table sizes and fragmentation for all databases - "
                            "[$yellow]Heed caution if you have a lot of tables![/$yellow]"
                        ),
                    },
                    "placeholder_4": {"human_key": "", "description": ""},
                    "c": {"human_key": "c", "description": "Clear all filters set"},
                    "f": {
                        "human_key": "f",
                        "description": "Filter threads by field(s)",
                    },
                    "E": {
                        "human_key": "E",
                        "description": "Export the processlist to a CSV file",
                    },
                    "k": {"human_key": "k", "description": "Kill thread(s)"},
                    "M": {"human_key": "M", "description": "Maximize a panel"},
                    "q": {"human_key": "q", "description": "Quit"},
                    "r": {"human_key": "r", "description": "Set the refresh interval"},
                    "R": {"human_key": "R", "description": "Reset all metrics"},
                    "space": {
                        "human_key": "space",
                        "description": "Force a manual refresh of all panels except replicas",
                    },
                }
            },
            ConnectionSource.proxysql: {
                "Commands": {
                    "1": {"human_key": "1", "description": "Toggle panel Dashboard"},
                    "2": {"human_key": "2", "description": "Toggle panel Processlist"},
                    "3": {
                        "human_key": "3",
                        "description": "Toggle panel Metric Graphs",
                    },
                    "4": {
                        "human_key": "4",
                        "description": "Toggle panel Hostgroup Summary",
                    },
                    "5": {"human_key": "5", "description": "Toggle panel Query Rules"},
                    "6": {
                        "human_key": "6",
                        "description": "Toggle panel Command Statistics",
                    },
                    "placeholder_1": {"human_key": "", "description": ""},
                    "grave_accent": {"human_key": "`", "description": "Open tab setup"},
                    "plus": {"human_key": "+", "description": "Create a new tab"},
                    "minus": {
                        "human_key": "-",
                        "description": "Remove the current tab",
                    },
                    "equals_sign": {
                        "human_key": "=",
                        "description": "Rename the current tab",
                    },
                    "D": {
                        "human_key": "D",
                        "description": "Disconnect from the tab's host",
                    },
                    "ctrl+a": {
                        "human_key": "ctrl+a",
                        "description": "Switch to the previous tab",
                    },
                    "ctrl+d": {
                        "human_key": "ctrl+d",
                        "description": "Switch to the next tab",
                    },
                    "placeholder_2": {"human_key": "", "description": ""},
                    "a": {
                        "human_key": "a",
                        "description": "Toggle additional processlist columns",
                    },
                    "i": {
                        "human_key": "i",
                        "description": "Toggle display of idle threads",
                    },
                    "p": {
                        "human_key": "p",
                        "description": "Toggle pause for refreshing of panels",
                    },
                    "s": {
                        "human_key": "s",
                        "description": "Toggle sorting for Age in Processlist panel",
                    },
                    "placeholder_3": {"human_key": "", "description": ""},
                    "e": {
                        "human_key": "e",
                        "description": "Display errors reported by backend servers during query execution",
                    },
                    "t": {
                        "human_key": "t",
                        "description": "Display details of a thread",
                    },
                    "u": {
                        "human_key": "u",
                        "description": "Display frontend users connected",
                    },
                    "v": {
                        "human_key": "v",
                        "description": "Display variables from SHOW GLOBAL VARIABLES",
                    },
                    "z": {
                        "human_key": "z",
                        "description": "Display all entries in the host cache",
                    },
                    "placeholder_4": {"human_key": "", "description": ""},
                    "c": {"human_key": "c", "description": "Clear all filters set"},
                    "f": {
                        "human_key": "f",
                        "description": "Filter threads by field(s)",
                    },
                    "E": {
                        "human_key": "E",
                        "description": "Export the processlist to a CSV file",
                    },
                    "k": {"human_key": "k", "description": "Kill thread(s)"},
                    "M": {"human_key": "M", "description": "Maximize a panel"},
                    "q": {"human_key": "q", "description": "Quit"},
                    "r": {"human_key": "r", "description": "Set the refresh interval"},
                    "R": {"human_key": "R", "description": "Reset all metrics"},
                    "space": {
                        "human_key": "space",
                        "description": "Force a manual refresh of all panels except replicas",
                    },
                },
                "Terminology": {
                    "FE": {"description": "Frontend"},
                    "BE": {"description": "Backend"},
                    "Conn": {"description": "Connection"},
                    "CP": {"description": "Connection Pool"},
                    "MP": {"description": "Multiplex"},
                },
            },
            "mysql_replay": {
                "Commands": {
                    "1": {"human_key": "1", "description": "Toggle panel Dashboard"},
                    "2": {"human_key": "2", "description": "Toggle panel Processlist"},
                    "3": {
                        "human_key": "3",
                        "description": "Toggle panel Metric Graphs",
                    },
                    "4": {
                        "human_key": "4",
                        "description": "Toggle panel Replication/Replicas",
                    },
                    "5": {
                        "human_key": "5",
                        "description": "Toggle panel Metadata Locks",
                    },
                    "7": {
                        "human_key": "7",
                        "description": "Toggle panel Performance Schema Metrics",
                    },
                    "8": {
                        "human_key": "8",
                        "description": "Toggle panel Statements Summary Metrics",
                    },
                    "placeholder_1": {"human_key": "", "description": ""},
                    "grave_accent": {"human_key": "`", "description": "Open tab setup"},
                    "plus": {"human_key": "+", "description": "Create a new tab"},
                    "minus": {
                        "human_key": "-",
                        "description": "Remove the current tab",
                    },
                    "equals_sign": {
                        "human_key": "=",
                        "description": "Rename the current tab",
                    },
                    "ctrl+a": {
                        "human_key": "ctrl+a",
                        "description": "Switch to the previous tab",
                    },
                    "ctrl+d": {
                        "human_key": "ctrl+d",
                        "description": "Switch to the next tab",
                    },
                    "placeholder_2": {"human_key": "", "description": ""},
                    "a": {
                        "human_key": "a",
                        "description": "Toggle additional processlist columns",
                    },
                    "A": {
                        "human_key": "A",
                        "description": "Toggle query digest text/sample text in Statements Summary panel",
                    },
                    "C": {
                        "human_key": "C",
                        "description": "Toggle display of concurrency threads with tickets in Processlist panel",
                    },
                    "T": {
                        "human_key": "T",
                        "description": "Toggle display of threads that only have an active transaction",
                    },
                    "s": {
                        "human_key": "s",
                        "description": "Toggle sorting for Age in Processlist panel",
                    },
                    "placeholder_3": {"human_key": "", "description": ""},
                    "t": {
                        "human_key": "t",
                        "description": "Display details of a thread",
                    },
                    "v": {
                        "human_key": "v",
                        "description": "Display global variables from SHOW GLOBAL VARIABLES",
                    },
                    "V": {
                        "human_key": "V",
                        "description": "Display global variables that changed during recording",
                    },
                    "placeholder_4": {"human_key": "", "description": ""},
                    "p": {"human_key": "p", "description": "Toggle pause of replay"},
                    "S": {
                        "human_key": "S",
                        "description": "Seek to a specific time in the replay",
                    },
                    "left_square_bracket": {
                        "human_key": "\\[",
                        "description": " Seek to previous refresh interval in the replay",
                    },
                    "right_square_bracket": {
                        "human_key": "]",
                        "description": "Seek to next refresh interval in the replay",
                    },
                    "placeholder_5": {"human_key": "", "description": ""},
                    "c": {"human_key": "c", "description": "Clear all filters set"},
                    "f": {
                        "human_key": "f",
                        "description": "Filter threads by field(s)",
                    },
                    "E": {
                        "human_key": "E",
                        "description": "Export the processlist to a CSV file",
                    },
                    "M": {"human_key": "M", "description": "Maximize a panel"},
                    "q": {"human_key": "q", "description": "Quit"},
                    "r": {"human_key": "r", "description": "Set the refresh interval"},
                }
            },
            "proxysql_replay": {
                "Commands": {
                    "1": {"human_key": "1", "description": "Toggle panel Dashboard"},
                    "2": {"human_key": "2", "description": "Toggle panel Processlist"},
                    "3": {
                        "human_key": "3",
                        "description": "Toggle panel Metric Graphs",
                    },
                    "4": {
                        "human_key": "4",
                        "description": "Toggle panel Hostgroup Summary",
                    },
                    "placeholder_1": {"human_key": "", "description": ""},
                    "grave_accent": {"human_key": "`", "description": "Open tab setup"},
                    "plus": {"human_key": "+", "description": "Create a new tab"},
                    "minus": {
                        "human_key": "-",
                        "description": "Remove the current tab",
                    },
                    "equals_sign": {
                        "human_key": "=",
                        "description": "Rename the current tab",
                    },
                    "ctrl+a": {
                        "human_key": "ctrl+a",
                        "description": "Switch to the previous tab",
                    },
                    "ctrl+d": {
                        "human_key": "ctrl+d",
                        "description": "Switch to the next tab",
                    },
                    "placeholder_2": {"human_key": "", "description": ""},
                    "a": {
                        "human_key": "a",
                        "description": "Toggle additional processlist columns",
                    },
                    "s": {
                        "human_key": "s",
                        "description": "Toggle sorting for Age in Processlist panel",
                    },
                    "placeholder_3": {"human_key": "", "description": ""},
                    "t": {
                        "human_key": "t",
                        "description": "Display details of a thread",
                    },
                    "v": {
                        "human_key": "v",
                        "description": "Display global variables from SHOW GLOBAL VARIABLES",
                    },
                    "V": {
                        "human_key": "V",
                        "description": "Display global variables that changed during recording",
                    },
                    "placeholder_4": {"human_key": "", "description": ""},
                    "p": {"human_key": "p", "description": "Toggle pause of replay"},
                    "S": {
                        "human_key": "S",
                        "description": "Seek to a specific time in the replay",
                    },
                    "left_square_bracket": {
                        "human_key": "[",
                        "description": "Seek to previous refresh interval in the replay",
                    },
                    "right_square_bracket": {
                        "human_key": "]",
                        "description": "Seek to next refresh interval in the replay",
                    },
                    "placeholder_5": {"human_key": "", "description": ""},
                    "c": {"human_key": "c", "description": "Clear all filters set"},
                    "f": {
                        "human_key": "f",
                        "description": "Filter threads by field(s)",
                    },
                    "E": {
                        "human_key": "E",
                        "description": "Export the processlist to a CSV file",
                    },
                    "M": {"human_key": "M", "description": "Maximize a panel"},
                    "q": {"human_key": "q", "description": "Quit"},
                    "r": {"human_key": "r", "description": "Set the refresh interval"},
                }
            },
        }

        # These are keys that we let go through no matter what
        self.exclude_keys = [
            "up",
            "down",
            "left",
            "right",
            "pageup",
            "pagedown",
            "home",
            "end",
            "tab",
            "enter",
            "grave_accent",
            "q",
            "question_mark",
            "plus",
            "minus",
            "equals_sign",
            "ctrl+a",
            "ctrl+d",
        ]

    def get_commands(self, replay_file: str, connection_source: ConnectionSource) -> dict[str, dict[str, str]]:
        if replay_file:
            key = {
                ConnectionSource.mysql: "mysql_replay",
                ConnectionSource.proxysql: "proxysql_replay",
            }.get(connection_source, connection_source)
        else:
            key = connection_source

        return self.command_keys.get(key, {}).get("Commands")



import re
from typing import Any

from dolphie.Modules.Functions import minify_query


class PerformanceSchemaMetrics:
    def __init__(self, query_data: list[dict[str, Any]], metric_name: str, metric_key: str):
        self.metric_name = metric_name
        self.metric_key = metric_key

        # These are integer columns that should be ignored for delta calculations
        self.ignore_int_columns = ["quantile_95", "quantile_99"]

        self.filtered_data: dict[str, dict[str, dict[str, int]]] = {}
        self.internal_data: dict[str, dict[str, dict[str, Any]]] = {
            row[self.metric_key]: {
                "event_name": row.get("EVENT_NAME"),
                "metrics": {
                    metric: {"total": value, "delta": 0, "delta_last_sample": 0}
                    for metric, value in row.items()
                    if isinstance(value, int) and metric not in self.ignore_int_columns
                },
            }
            for row in query_data
            if row[self.metric_key] is not None
        }

        self.table_pattern = re.compile(r"([^/]+)/([^/]+)\.(frm|ibd|MYD|MYI|CSM|CSV|par)$")
        self.undo_logs_pattern = re.compile(r"undo_\d+$")

        self.events_to_combine = {
            "wait/io/file/innodb/innodb_temp_file": "Temporary files",
            "wait/io/file/sql/binlog": "Binary logs",
            "wait/io/file/sql/relaylog": "Relay logs",
            "wait/io/file/sql/io_cache": "IO cache",
            "wait/io/file/innodb/innodb_dblwr_file": "Doublewrite buffer",
            "wait/io/file/innodb/innodb_log_file": "InnoDB redo logs",
            "wait/io/file/sql/hash_join": "Hash joins",
        }

    def update_internal_data(self, query_data: list[dict[str, int]]):
        # Track instances and remove missing ones
        current_instance_names = {row[self.metric_key] for row in query_data if row[self.metric_key] is not None}
        instances_to_remove = set(self.internal_data) - current_instance_names

        # Process current query data
        for row in query_data:
            instance_name = row[self.metric_key]
            # Skip rows where the metric key is None (orjson cannot serialize None as dict keys)
            if instance_name is None:
                continue
            metrics = {
                metric: value
                for metric, value in row.items()
                if isinstance(value, int) and metric not in self.ignore_int_columns
            }

            # Initialize instance in internal_data if not present
            if instance_name not in self.internal_data:
                self.internal_data[instance_name] = {
                    "event_name": row.get("EVENT_NAME"),
                    "metrics": {
                        metric: {"total": value, "delta": 0, "delta_last_sample": 0}
                        for metric, value in metrics.items()
                    },
                }

            deltas_changed = False
            all_deltas_zero = True

            # Update deltas for each metric
            for metric, current_value in metrics.items():
                metric_data = self.internal_data[instance_name]["metrics"][metric]
                initial_value = metric_data["total"]
                delta = current_value - initial_value
                previous_delta = metric_data["delta_last_sample"]

                metric_data["total"] = current_value
                metric_data["delta_last_sample"] = delta
                if delta > 0:
                    metric_data["delta"] += delta
                    deltas_changed = True
                elif previous_delta > 0:
                    deltas_changed = True

                if metric_data["delta"] > 0:
                    all_deltas_zero = False

            # Update filtered_data with new values if deltas changed or instance is new
            if deltas_changed or instance_name not in self.filtered_data:
                self.filtered_data[instance_name] = {}

                for metric, values in self.internal_data[instance_name]["metrics"].items():
                    # Update total with the new value (whether or not delta is positive)
                    total = values["total"]

                    # Only add delta if it's greater than 0
                    delta = values["delta"] if values["delta"] > 0 else 0
                    delta_last_sample = values["delta_last_sample"] if values["delta_last_sample"] > 0 else 0

                    # Only include the metric in filtered_data if it has a delta greater than 0
                    if delta > 0:
                        self.filtered_data[instance_name][metric] = {
                            "t": total,
                            "d": delta,
                            "d_last_sample": delta_last_sample,
                        }
                    else:
                        self.filtered_data[instance_name][metric] = {"t": total}

                    if (
                        self.metric_name == "statements_summary"
                        and "schema_name" not in self.filtered_data[instance_name]
                    ):
                        self.filtered_data[instance_name]["schema_name"] = row.get("schema_name")
                        self.filtered_data[instance_name]["digest_text"] = minify_query(row.get("digest_text"))
                        self.filtered_data[instance_name]["query_sample_text"] = minify_query(
                            row.get("query_sample_text")
                        )
                        self.filtered_data[instance_name]["quantile_95"] = row.get("quantile_95")
                        self.filtered_data[instance_name]["quantile_99"] = row.get("quantile_99")

            if all_deltas_zero:
                self.filtered_data.pop(instance_name, None)

        # Remove instances no longer in the query data
        for instance_name in instances_to_remove:
            del self.internal_data[instance_name]

        if self.metric_name == "file_io":
            self.aggregate_and_combine_data()

    def aggregate_and_combine_data(self):
        combined_results = {}

        # Aggregate deltas for combined events and instances matching the regex
        for instance_name, instance_data in self.internal_data.items():
            event_name = instance_data["event_name"]

            # Determine the target name based on instance name pattern or specific event name
            if self.undo_logs_pattern.search(instance_name):
                target_name = "Undo Logs"
            elif event_name in self.events_to_combine:
                target_name = self.events_to_combine[event_name]
            else:
                continue  # Skip if it doesn't match any pattern or event to combine

            # Remove original instance from filtered_data if it exists
            self.filtered_data.pop(instance_name, None)

            # Initialize target in combined results if not already present
            target_metrics = combined_results.setdefault(target_name, {})

            # Accumulate metrics for each matched or combined event
            for metric_name, metric_data in instance_data["metrics"].items():
                combined_metric = target_metrics.setdefault(metric_name, {"total": 0, "delta": 0})
                combined_metric["total"] += metric_data["total"]
                combined_metric["delta"] += metric_data["delta"]

        # Update filtered_data with combined results only if SUM_TIMER_WAIT delta > 0
        for combined_name, combined_metrics in combined_results.items():
            # Skip if SUM_TIMER_WAIT delta is 0
            if combined_metrics.get("SUM_TIMER_WAIT", {}).get("delta", 0) > 0:
                self.filtered_data[combined_name] = {
                    metric_name: {"t": combined_data["total"], "d": combined_data["delta"]}
                    for metric_name, combined_data in combined_metrics.items()
                }

        # Clean up filtered_data by removing instances with SUM_TIMER_WAIT delta of 0
        self.filtered_data = {
            instance_name: instance_metrics
            for instance_name, instance_metrics in self.filtered_data.items()
            if instance_metrics.get("SUM_TIMER_WAIT", {}).get("d", 0) != 0
        }



import re
import string
import time
from ssl import SSLError

import pymysql
from loguru import logger
from textual.app import App

from dolphie.DataTypes import ConnectionSource
from dolphie.Modules.ManualException import ManualException
from dolphie.Modules.Queries import MySQLQueries, ProxySQLQueries


class Database:
    def __init__(
        self,
        app: App,
        host: str,
        user: str,
        password: str,
        socket: str,
        port: int,
        ssl: str,
        save_connection_id: bool = True,
        auto_connect: bool = True,
        daemon_mode: bool = False,
    ):
        self.app = app
        self.host = host
        self.user = user
        self.password = password
        self.socket = socket
        self.port = port
        self.ssl = ssl
        self.save_connection_id = save_connection_id
        self.daemon_mode = daemon_mode

        self._PRIVILEGE_ERROR_CODES = {
            1227,  # Access denied; SUPER privilege
            1370,  # execute command denied; REPLICATION CLIENT privilege
            1044,  # Access denied for user to database
            1142,  # command denied to user
            1143,  # column command denied to user
        }

        self.connection: pymysql.Connection = None
        self.connection_id: int = None
        self.source: ConnectionSource = None
        self.is_running_query: bool = False
        self.has_connected: bool = False
        self.last_execute_successful: bool = False
        self.privilege_errors_notified: set = (
            set()
        )  # Track queries that have already shown privilege error notifications

        # Pre-compile regex pattern to filter non-printable characters
        self.non_printable_regex = re.compile(f"[^{re.escape(string.printable)}]")

        if daemon_mode:
            self.max_reconnect_attempts: int = 999999999
        else:
            self.max_reconnect_attempts: int = 3

        if auto_connect:
            self.connect()

    def connect(self, reconnect_attempt: bool = False):
        try:
            self.connection = pymysql.connect(
                host=self.host,
                user=self.user,
                passwd=self.password,
                unix_socket=self.socket,
                port=int(self.port),
                use_unicode=False,
                ssl=self.ssl,
                autocommit=True,
                connect_timeout=5,
                program_name="Dolphie",
            )
            self.cursor = self.connection.cursor(pymysql.cursors.DictCursor)

            # If the query is successful, then the connection is to ProxySQL
            try:
                self.cursor.execute("SELECT @@admin-version")
                self.source = ConnectionSource.proxysql
            except Exception:
                self.source = ConnectionSource.mysql

            # Get connection ID for processlist filtering
            if self.save_connection_id:
                self.connection_id = self.connection.thread_id()

            # We don't want any SQL modes to be set to avoid unexpected behavior between MySQL & MariaDB
            if self.source == ConnectionSource.mysql:
                self.execute("SET SESSION sql_mode=''")

            logger.info(f"Connected to {self.source} with Process ID {self.connection_id}")
            self.has_connected = True
        except pymysql.Error as e:
            if reconnect_attempt:
                logger.error(f"Failed to reconnect to {self.source}: {e.args[1]}")
                escaped_error_message = e.args[1].replace("[", "\\[")
                self.app.notify(
                    (
                        f"[$b_light_blue]{self.host}:{self.port}[/$b_light_blue]: "
                        f"Failed to reconnect to MySQL: {escaped_error_message}"
                    ),
                    title="MySQL Reconnection Failed",
                    severity="error",
                    timeout=10,
                )
            else:
                if len(e.args) == 1:
                    raise ManualException(e.args[0]) from e
                else:
                    raise ManualException(e.args[1]) from e
        except FileNotFoundError as e:  # Catch SSL file path errors
            raise ManualException("SSL certificate file path isn't valid!") from e
        except SSLError as e:
            raise ManualException(f"SSL error: {e}") from e

    def close(self):
        if self.is_connected():
            self.connection.close()

    def is_connected(self) -> bool:
        return self.connection and self.connection.open

    def _process_row(self, row):
        return {field: self._decode_value(value) for field, value in row.items()}

    def _decode_value(self, value):
        if isinstance(value, (bytes, bytearray)):
            # First attempt: UTF-8
            try:
                decoded_value = value.decode("utf-8")
            except UnicodeDecodeError:
                # Second attempt: Latin-1
                try:
                    decoded_value = value.decode("latin-1")
                except UnicodeDecodeError:
                    # Fallback: Hex representation
                    return f"/* Failed to decode query, returning hex: {value.hex()} */"

            # Skip regex substitution for pure ASCII values (vast majority of MySQL system data)
            if decoded_value.isascii():
                return decoded_value

            return self.non_printable_regex.sub("?", decoded_value)

        return value

    def fetchall(self):
        if not self.is_connected() or not self.last_execute_successful:
            return []

        rows = self.cursor.fetchall()
        return [self._process_row(row) for row in rows] if rows else []

    def fetchone(self):
        if not self.is_connected() or not self.last_execute_successful:
            return {}

        row = self.cursor.fetchone()
        return self._process_row(row) if row else {}

    def fetch_value_from_field(self, query, field=None, values=None):
        if not self.is_connected():
            return None

        self.execute(query, values)
        data = self.cursor.fetchone()

        if not data:
            return None

        field = field or next(iter(data))  # Use field if provided, otherwise get first field
        value = data.get(field)
        return self._decode_value(value)

    def fetch_status_and_variables(self, command):
        if not self.is_connected():
            return None

        self.execute(
            getattr(ProxySQLQueries, command)
            if self.source == ConnectionSource.proxysql
            else getattr(MySQLQueries, command)
        )
        data = self.fetchall()

        if command in {"status", "variables", "mysql_stats"}:
            return {
                row["Variable_name"]: int(row["Value"]) if row["Value"].isnumeric() else row["Value"] for row in data
            }
        elif command == "innodb_metrics":
            return {row["NAME"]: int(row["COUNT"]) for row in data}

    def execute(self, query, values=None, ignore_error=False):
        if not self.is_connected():
            self.last_execute_successful = False
            return None

        if self.is_running_query:
            self.app.notify(
                "Another query is already running, please repeat action",
                title="Unable to run multiple queries at the same time",
                severity="error",
                timeout=10,
            )
            self.last_execute_successful = False
            return None

        # Prefix all queries with Dolphie so they can be easily identified in the processlist
        if self.source != ConnectionSource.proxysql:
            query = "/* Dolphie */ " + query

        # Check if this query has already failed with a privilege error - skip execution to save database call
        raw_query = query.replace("/* Dolphie */ ", "")
        if raw_query in self.privilege_errors_notified:
            self.last_execute_successful = False
            return None

        for attempt_number in range(self.max_reconnect_attempts):
            self.is_running_query = True
            error_message = None

            try:
                rows = self.cursor.execute(query, values)
                self.is_running_query = False
                self.last_execute_successful = True

                return rows
            except AttributeError:
                # If the cursor is not defined, reconnect and try again
                self.is_running_query = False
                self.last_execute_successful = False

                self.close()
                self.connect()

                time.sleep(1)
            except pymysql.Error as e:
                self.is_running_query = False
                self.last_execute_successful = False

                if len(e.args) == 1:
                    error_code = e.args[0]
                else:
                    error_code = e.args[0]
                    if e.args[1]:
                        error_message = e.args[1]

                # Check if this is a privilege error - silently return None without raising exception
                if error_code in self._PRIVILEGE_ERROR_CODES:
                    # Show notification only the first time this query fails with privilege error
                    if raw_query not in self.privilege_errors_notified:
                        self.privilege_errors_notified.add(raw_query)

                        logger.warning(
                            f"Privilege error (code {error_code}): {error_message}. "
                            f"Query: {raw_query}. "
                            f"This query will be skipped and stats for this feature won't be available."
                        )

                        # Escape [ and ] characters in the error message and query
                        escaped_error_message = error_message.replace("[", "\\[") if error_message else "Access denied"
                        escaped_query = raw_query.replace("[", "\\[")

                        self.app.notify(
                            f"[$b_highlight]{self.host}:{self.port}[/$b_highlight]: [dim]{error_code}: "
                            f"{escaped_error_message}[/dim]\nQuery: [$b_light_blue]{escaped_query}[/$b_light_blue]\n"
                            "Stats for this feature won't be available.",
                            title="Insufficient Privileges",
                            severity="warning",
                            timeout=9,
                        )

                    return None

                # If ignore_error is set, return None for any error
                if ignore_error:
                    return None

                # Handle lost connection errors
                if error_code in (0, 2006, 2013, 2055):
                    # 0: Not connected to MySQL
                    # 2006: MySQL server has gone away
                    # 2013: Lost connection to MySQL server during query
                    # 2055: Lost connection to MySQL server at hostname

                    if error_message:
                        logger.error(
                            f"{self.source} has lost its connection: {error_message}, attempting to reconnect..."
                        )
                        # Escape [ and ] characters in the error message
                        escaped_error_message = error_message.replace("[", "\\[")
                        self.app.notify(
                            f"[$b_light_blue]{self.host}:{self.port}[/$b_light_blue]: {escaped_error_message}",
                            title="MySQL Connection Lost",
                            severity="error",
                            timeout=10,
                        )

                    self.close()
                    self.connect(reconnect_attempt=True)

                    if not self.connection.open:
                        # Exponential backoff
                        time.sleep(min(1 * (2**attempt_number), 20))  # Cap the wait time at 20 seconds

                        # Skip the rest of the loop
                        continue

                    self.app.notify(
                        f"[$b_light_blue]{self.host}:{self.port}[/$b_light_blue]: Successfully reconnected",
                        title="MySQL Connection Created",
                        severity="success",
                        timeout=10,
                    )

                    # Retry the query
                    return self.execute(query, values)
                else:
                    raise ManualException(error_message, query=query, code=error_code) from e

        if not self.connection.open:
            raise ManualException(
                f"Failed to reconnect to {self.source} after {self.max_reconnect_attempts} attempts",
                query=query,
                code=error_code,
            )
